import os
import sys
import time
import logging
import argparse
import pandas as pd

sys.path.insert(0, '..')
from GetLineage import MatchVariant, TrackLineage
from FilesIO import CommonData

common_data = CommonData()
test_path = os.path.split(os.path.realpath(__file__))[0] + '/TestData/'
out_path = test_path + 'TestOut/'
if not os.path.exists(out_path):
    os.makedirs(out_path)


def test_classify(input, format, build, sample):

    cutoff = None
    snp_only = False
    appro = False
    chip = False
    mutation = True
    output_set = [out_path + 'Test.hapresult.hg',
                  out_path + 'Test.lineageresult.hg',
                  out_path + 'Test.mutationresult.txt']

    # read ISOGG haplogroup reference data
    ref_info, ref_ind_info = common_data.read_ref_info(snp_only, appro)
    # filter samples
    from ProcessData import ReadFilterData
    read_filter = ReadFilterData.ReadFilterData(input, sample)
    if format == 'bam':
        data_info, inds = read_filter.read_bam(ref_info, build, 'snp')
    else:
        data_info, inds = read_filter.read_filter_variant(cutoff, format)
    # match input data to ISOGG reference
    match = MatchVariant.MatchVariant(build)
    matched_info, header_cut = match.match_variant_snp(ref_info, data_info, format)

    # classify haplogroups and build lineage track
    tracker = TrackLineage.TrackLineage(matched_info, ref_ind_info, build, format, chip)
    hap_data = tracker.classify_haplogroup(inds, matched_info, ref_info, output_set, mutation, header_cut)


def test_cluster(hg_data, pop_data, method):

    from ClusterHaplogroup import ClusterAnalysis

    # set parameters
    level = 3
    freq = True
    if method == 'pca':
        output_set = [out_path + 'Test.PCA.fig.pdf',
                      out_path + 'Test.PCA.eigenvectors.txt',
                      out_path + 'Test.PCA.eigenval.txt',
                      out_path + 'Test.level.hg',
                      out_path + 'Test.freq.txt']
    else:
        output_set = [out_path + 'Test.MDS.fig.pdf',
                      out_path + 'Test.MDS.fst.txt',
                      out_path + 'Test.MDS.embedding.txt',
                      out_path + 'Test.level.hg',
                      out_path + 'Test.freq.txt']

    # start cluserting analysis
    cluster = ClusterAnalysis()
    hg_data = cluster.classify_hap_data(level, hg_data, output_set[3])
    cluster.output_cluster(method, freq, output_set, hg_data, pop_data)


def test_phylo(hg_data, pop_data, seq_file):

    from PhyloHaplogroup import PhyloHaplogroup

    align = 'mp'
    format = 'vcf'
    tree = 'newick'
    layout = 'rectangular'

    # set of output files
    output_set = [out_path + 'Test.phylo.fig.pdf',
                  out_path + 'Test.tree.nwk']

    # phylo analysis
    phylo = PhyloHaplogroup(hg_data, output_set)
    ids, seqs = phylo.read_seq_file(seq_file, format)
    hap_df, base, trunk_length, max_length, samples = phylo.get_base_tree(ids)
    base = phylo.phylo_internal(ids, seqs, base, samples, align)
    pop_info = pd.merge(hap_df, pop_data, on='SampleID')
    pop_info = pop_info.set_index('SampleID')
    phylo.output_tree(base, tree)
    #phylo.output_tree_figure(pop_info, base, trunk_length, max_length, samples, layout)


def test_genostr():

    import pysam


def test_net(hap_data, pop_data):

    from HaploNetwork import HaploNetwork

    class NetArgs(object):

        gap = 'allele'
        filter = 0.7
        treetype = 'mjn'
        search = 0
        alt = 'hide'
        mediansize = 0
        fdi = True
        euqalsize = False

    arguments = NetArgs()
    gap = 'allele'
    label = 'Population'
    path = out_path + 'Test'

    # calculate network and visualize result
    net = HaploNetwork(gap, path)
    net.calculate_and_output(hap_data, pop_data, arguments, label)


def test_stat(hap_data, pop_data):

    from StatisticsAnalysis import StatisticsAnalysis, predict_haplogroup

    class StatArgs(object):

        hd = True
        mpd = True
        gd = False
        amova = 'fst'

    arguments = StatArgs()
    path = out_path + 'Test'
    label = 'Population'
    stat = StatisticsAnalysis(label, hap_data, pop_data)
    stat.calculate_and_output(arguments, path)
    predict_haplogroup(hap_data, path)


def test_time(seq_file, tree_file):

    from EstimateTime import EstimateTime

    class TimeArgs(object):

        tree = tree_file
        seq = seq_file
        format = 'vcf'
        filter = 0.1
        mcmc = 2000
        rate = 7.6e-10
        model = 'hky85'
        calibration = 'O2a:10000-13000'
        auto = False

    arguments = TimeArgs()
    path = out_path + 'Test'

    # estimate divergence time
    estimate = EstimateTime(path)
    estimate.run_mcmc(arguments)

def test_tmrca(input_file, format, pop_file):

    from CommonAncestor import CommonAncestor

    class TmrcaArgs(object):

        if format == 'net':
            net = input_file
            matrix = None
            population = None
        else:
            net = None
            matrix = input_file
            population = pop_file
        rate = 6e-3
        generation = 30
        ad = ['NA18530:NA18543,NA18544']

    arguments = TmrcaArgs()
    path = out_path + 'Test'

    estimate = CommonAncestor()
    estimate.run_tmrca(arguments, path)


def main():

    from FilesIO import check_hap_input, check_population, check_overlap, set_haplotype_index

    vcf_file = test_path + '/KGP_sample7.vcf'
    tree_file = test_path + '/KGP_sample7.tree.nwk'
    net_file = test_path + '/KGP_sample7.net'

    hg_file = test_path + '/KGP_hgs.hg'
    hg_data = check_hap_input(hg_file, 'haplogroup')

    pop_file = test_path + '/KGP_pops.pop'
    pop_data = check_population(pop_file)

    hap_file = test_path + '/KGP_O2_sample7.matrix'
    hap_data = check_hap_input(hap_file, 'haplotype', 'matrix')

    hap_data, pop_data2 = check_overlap(hap_data, pop_data)
    hap_data = set_haplotype_index(hap_data, 'matrix')

    print('[Y-LineageTracker] Testing classify command...')
    try:
        test_classify(vcf_file, 'vcf', 37, None)
        print('[Y-LineageTracker] Successfully tested classify command')
    except:
        print('[Y-LineageTracker] Failed in testing classify command')

    print('[Y-LineageTracker] Testing cluster command...')
    try:
        test_cluster(hg_data, pop_data, 'pca')
        test_cluster(hg_data, pop_data, 'mds')
        print('[Y-LineageTracker] Successfully tested clster command')
    except:
        print('[Y-LineageTracker] Failed in testing cluster command')

    print('[Y-LineageTracker] Testing phylo command...')
    try:
        test_phylo(hg_data, pop_data, vcf_file)
        print('[Y-LineageTracker] Successfully tested phylo command')
    except:
        print('[Y-LineageTracker] Failed in testing phylo command')

    print('[Y-LineageTracker] Testing genostr command...')
    try:
        test_genostr()
        print('[Y-LineageTracker] Successfully tested genostr command')
    except:
        print('[Y-LineageTracker] Failed in testing genostr command')

    print('[Y-LineageTracker] Testing net command...')
    try:
        test_net(hap_data, pop_data2)
        print('[Y-LineageTracker] Successfully tested net command')
    except:
        print('[Y-LineageTracker] Failed in testing net command')

    print('[Y-LineageTracker] Testing stat command...')
    try:
        test_stat(hap_data, pop_data2)
        print('[Y-LineageTracker] Successfully tested stat command')
    except:
        print('[Y-LineageTracker] Failed in testing stat command')

    print('[Y-LineageTracker] Testing time command...')
    try:
        test_time(vcf_file, tree_file)
        print('[Y-LineageTracker] Successfully tested time command')
    except:
        print('[Y-LineageTracker] Failed in testing time command')

    print('[Y-LineageTracker] Testing tmrca command...')
    try:
        test_tmrca(hap_file, 'hap', pop_file)
        test_tmrca(net_file, 'net', pop_file)
        print('[Y-LineageTracker] Successfully tested tmrca command')
    except:
        print('[Y-LineageTracker] Failed in testing tmrca command')


if __name__ == '__main__':
    main()



import sys
import time
import random
import logging
import argparse
import pandas as pd
import numpy as np

from itertools import chain, repeat
from collections import Counter
from GetConfig import getConfig


config = getConfig()


'''
Description:
This module is used for statistical analysis.
Y haplotype data is required for analysis.
It can calculate commonly used statistics in population genetics.
HD: haplotype diversity
MPD: mean pairwise differences
GD: genetic distance
AMOVA: analysis of molecular variance
predict: predict haplogroup from Y-STR haplotype
'''


def stat_parser():

    parser = argparse.ArgumentParser('stat', description='(c) Y-LineageTracker: Statistical analysis')
    # function used for statistical analysis
    parser.add_argument('stat',
                        help='Perform statistical analysis from Y haplotype data.')
    # required, input file, alignment sequence or matrix
    input = parser.add_mutually_exclusive_group(required=True)
    input.add_argument('--seq',
                        type=str,
                        action='store',
                        help='seq: Haplotype data in sequence alignment format.')
    input.add_argument('--matrix',
                        type=str,
                        action='store',
                        help='matrix: Y-STR haplotype data in matrix format.')
    # optional, format of sequence alignment
    parser.add_argument('--seq-format',
                        required=False,
                        type=str,
                        dest='format',
                        action='store',
                        choices=['fasta', 'phylip', 'nexus', 'meg', 'vcf'],
                        help='seq-format: The format of sequence file. \
                              This option is only required for the sequence alignment file.')
    # population file, required for calculating statistics, not required for prediction
    parser.add_argument('-p', '--population',
                        required=False,
                        type=str,
                        action='store',
                        help='population: A file containing sample ID and population information of each individual. \
                              The population file is required in the analysis of calculating statistics, but not in the haplogroup prediction')
    # optional, calculate genetic distance
    parser.add_argument('--gd',
                        required=False,
                        type=str,
                        action='store',
                        choices=['fst', 'rst', 'Ngst', 'NCgst', 'Hgst', 'ibs'],
                        help='gd: Calculate genetic distance based on the selected statistics.')
    # optional, perform analysis of molecular variance
    parser.add_argument('--amova',
                        required=False,
                        type=str,
                        action='store',
                        nargs='?',
                        choices=['fst', 'rst'],
                        help='amova: Get the AMOVA table and calculate Fst or Rst of pairwise populations based on AMOVA and give the p values. \
                              The default is fst.')
    # optional, calculate haplotype diversity
    parser.add_argument('--hd',
                        required=False,
                        action='store_true',
                        help='hd: Calculate haplotype diversity of each population.')
    # optional, calculate mean pairwise differences
    parser.add_argument('--mpd',
                        required=False,
                        action='store_true',
                        help='mpd: calculate mean pairwise differences within and between populations.')
    # optional, perform prediction anlaysis from Y-STR haplotype
    parser.add_argument('--predict',
                        required=False,
                        action='store_true',
                        help='predict: predict possible NRY haplogroups and give the probability of each haplogroup from Y-STR haplotype data by Bayesian approach. \
                              This analysis only supports Y-STR haplotype data.')
    # optional, the prefix of output
    parser.add_argument('-o', '--output',
                        required=False,
                        type=str,
                        action='store',
                        help='output: The prefix of output files.')

    args = parser.parse_args()

    return args


# print program information and write to log file
def set_log(args_log, log_file):

    logger = logging.getLogger()
    logger.setLevel(level=logging.INFO)

    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] - [%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    logger.addHandler(handler)
    logger.addHandler(console)

    log_info = ['[Y-LineageTracker] [Stat]',
                '[Y-LineageTracker] Run Date: ' + time.asctime(time.localtime(time.time())),
                '[Y-LineageTracker] Haplotype File: %s' % args_log.seq,
                '[Y-LineageTracker] Haplotype File Format: %s' % args_log.format]

    if args_log.mpd:
        log_info.append('[Y-LineageTracker] Calculate mean pairwise differences')
    if args_log.hd:
        log_info.append('[Y-LineageTracker] Calculate haplotype diversity')
    if args_log.amova:
        log_info.append('[Y-LineageTracker] Calculate %s based on AMOVA' % args_log.amova)
    if args_log.predict:
        log_info.append('[Y-LineageTracker] Predict haplogroup from Y-STR data')

    print('\n')
    for i in log_info:
        logger.info(i)


# check input command
def check_commands(arguments):

    if arguments.mpd or arguments.hd or arguments.amova or arguments.gd:
        if not arguments.population:
            print('[Y-LineageTracker] Population file required for specified analysis')
            sys.exit()
    else:
        if not arguments.predict:
            print('[Y-LineageTracker] No command for analysis')
            sys.exit()


# check amova command
def check_amova(amova, arguments):

    if amova == None:
        if '--amova' in arguments:
            amova = 'fst'
    else:
        amova = amova

    return amova


class StatisticsAnalysis(object):
    '''
    This class include commonly used statistical method for Y-chromosome analysis:
    1. HD: haplotype diversity
    2. MPD: mean pairwise differences
    3. GD: genetic distance
    4. AMOVA: analysis of molecular analysis
    for analysis 2-4, p value will calculated by permutation
    '''

    def __init__(self, label, haplotype_data, population_data):

        self.logger = logging.getLogger()
        self.label = label
        if label == 'Population':
            self.label_num = 1
        else:
            self.label_num = 2
        self.skip_letters = config.get('Statistics', 'SkipLetters').split(',')
        self.haplotype_array = np.array(haplotype_data)
        self.haplotype_array_with_inds = haplotype_data.reset_index().to_numpy()
        self.population_array = np.array(population_data)
        self.header = haplotype_data.columns.tolist()
        self.pops = sorted(list(set(population_data[label])))

    # calculate distance between two sequence
    def _pairwise_distance(self, seq1, seq2, type):

        if type == 'num_of_diff_alleles':
            distance = sum(l1 != l2 for l1, l2 in zip(seq1, seq2) if l1 not in self.skip_letters and l2 not in self.skip_letters)
        elif type == 'sum_of_squared_diff':
            distance = sum((int(l1)-int(l2))**2 for l1, l2 in zip(seq1, seq2) if l1 not in self.skip_letters and l2 not in self.skip_letters)

        return distance

    # get haplotype data of a population
    def _get_population_data(self, pop, population_data=None):

        # string type: one population, list type: two populations or more
        if isinstance(population_data, np.ndarray):
            population_data = population_data
        else:
            population_data = self.population_array
        if isinstance(pop, str):
            pop_inds = population_data[population_data[:, self.label_num]==pop][:, 0]
        elif isinstance(pop, list):
            pop_inds = population_data[list(map(lambda x: x in pop, population_data[:, self.label_num]))][:, 0]
        pop_haplotype_data = self.haplotype_array[list(map(lambda x: x in pop_inds, self.haplotype_array_with_inds[:, 0]))]

        return pop_haplotype_data

    # calculate haplotype frequency of a population
    def _calculate_haplotype_freq(self, pop_haplotype_data):

        pop_all_haps = [tuple(i) for i in pop_haplotype_data] # get all haplotypes
        pop_haps = sorted(list(set(pop_all_haps))) # get haplotypes without duplication
        hap_freq = [pop_all_haps.count(i) / len(pop_all_haps) for i in pop_haps] # calculate freq of each haplotype for the population

        return pop_haps, hap_freq

    # calculate allele frequency of a population
    def _calculate_allele_freq(self, pops, population_data, pops_haplotype_data):

        # get allele frequency, sample size of each population
        pops_count = []
        sample_sizes = []
        max_len = max([len(set(pops_haplotype_data[:, i])) for i in range(len(self.header))])
        for pop in pops:
            pop_haplotype_data = self._get_population_data(pop, population_data)
            single_pop_count = []
            sample_sizes.append(len(pop_haplotype_data))
            for allele in range(len(self.header)):
                all_alleles = sorted(list(set(pops_haplotype_data[:, allele])))
                allele_count = [list(pop_haplotype_data[:, allele]).count(i) for i in all_alleles]
                if len(allele_count) < max_len:
                    diff = max_len - len(allele_count)
                    allele_count += [np.nan] * diff
                single_pop_count.append(allele_count)
            pops_count.append(single_pop_count)
        pops_count = np.array(pops_count)
        sample_sizes = np.array(sample_sizes)

        return pops_count, sample_sizes

    # calculate haplotype diversity of a population
    def _haplotype_diversity(self, pop, type, allele=None):

        pop_haplotype_data = self._get_population_data(pop)
        n = len(pop_haplotype_data)
        if type == 'single':
            hap_freq = (np.unique(pop_haplotype_data[:, allele], return_counts=True)[1]/len(pop_haplotype_data[:, allele]))
        elif type == 'haplotype':
            pop_haps, hap_freq = self. _calculate_haplotype_freq(pop_haplotype_data)

        if n == 1:
            return 'NA', 'NA'

        # calculate single Y-STR and Y-STR haplotype diversity
        sub = 0
        for i in hap_freq:
            sub += i**2
        diversity = (n/(n-1)) * (1-sub)

        # calculate var and sd
        sub1 = 0
        sub2 = 0
        for i in hap_freq:
            sub1 += i**3
            sub2 += i**2

        v1 = 2.0 / (n*(n-1))
        v2 = 2.0 * (n-2) * (sub1-sub2**2)

        var = v1 * (v2+sub2-sub2**2)
        sd = var**0.5

        return diversity, sd

    # get haplotype diversity of all populations
    def _calculate_haplotype_diversity(self):

        hd_df = pd.DataFrame(index=self.pops, columns=['Haplotype']+self.header)
        for pop in self.pops:
            for n, allele in enumerate(self.header):
                diversity, sd = self._haplotype_diversity(pop, 'single', n)
                hd_df.at[pop, allele] = str(round(diversity, 5))+'±'+str(round(sd, 5))
            diversity, sd = self._haplotype_diversity(pop, 'haplotype')
            if diversity == 'NA':
                hd_df.at[pop, 'Haplotype'] = 'NA'
            else:
                hd_df.at[pop, 'Haplotype'] = str(round(diversity, 5))+'±'+str(round(sd, 5))

        return hd_df

    # calculate mean pairwise differences of between two populations
    def _between_mean_pairwise_differences(self, pop1, pop2):

        pop1_data = self._get_population_data(pop1)
        pop2_data = self._get_population_data(pop2)
        n = len(pop1_data) + len(pop2_data)
        pop1_haps, hap1_freq = self._calculate_haplotype_freq(pop1_data)
        pop2_haps, hap2_freq = self._calculate_haplotype_freq(pop2_data)

        pi = 0
        for i in range(len(pop1_haps)):
            Sub = 0
            for j in range(len(pop2_haps)):
                hap1 = pop1_haps[i]
                hap2 = pop2_haps[j]
                d = self._pairwise_distance(hap1, hap2, 'num_of_diff_alleles')
                p_i = hap1_freq[i]
                p_j = hap2_freq[j]
                Sub += p_i*p_j*d
            pi += Sub

        if n > 6:
            # calculate var ans sd
            sub1 = 3 * n * (n+1)
            sub2 = 2 * (n**2+n+3)
            sub3 = 11 * (n**2-7*n+6)

            var = float(sub1*pi+sub2*pi**2) / sub3
            sd = var**0.5
        else:
            sd = 'NA'

        return pi, sd

    # calculate mean pairwise difference within a population
    def _within_mean_pairwise_differences(self, pop):

        pop_haplotype_data = self._get_population_data(pop)
        n = len(pop_haplotype_data)
        pop_haps, hap_freq = self. _calculate_haplotype_freq(pop_haplotype_data)

        # n: sample size
        # hap_freq: frequence of haplotypes of two population
        if round(sum(hap_freq), 1) != 1.0:
            return 'NA', 'NA'
        if n <= 2:
            return 'NA', 'NA'

        pi = 0
        k = len(pop_haps)
        for i in range(k):
            Sub = 0
            for j in range(k):
                hap1 = pop_haps[i]
                hap2 = pop_haps[j]
                d = self._pairwise_distance(hap1, hap2, 'num_of_diff_alleles')
                p_i = hap_freq[i]
                p_j = hap_freq[j]
                Sub += p_i*p_j*d
            pi += Sub
        pi = (float(n)/(n-1)) * pi

        if n > 6:
            # calculate var ans sd
            sub1 = 3 * n * (n+1)
            sub2 = 2 * (n**2+n+3)
            sub3 = 11 * (n**2-7*n+6)
            var = float(sub1*pi+sub2*pi**2) / sub3
            sd = var**0.5
        else:
            sd = 'NA'

        return pi, sd

    # get mean pairwise difference of all populations
    def _calculate_mean_pairwise_differences(self):

        mpd_df = pd.DataFrame(index=self.pops, columns=self.pops)
        pop_num = 0
        for pop1 in self.pops:
            for pop2 in self.pops[pop_num:]:
                if pop1 == pop2:
                    pi, sd = self._within_mean_pairwise_differences(pop1)
                else:
                    pi, sd = self._between_mean_pairwise_differences(pop1, pop2)
                if pi == 'NA':
                    mpd_df.at[pop2, pop1] = 'NA'
                else:
                    if sd == 'NA':
                        mpd_df.at[pop2, pop1] = str(round(pi, 5))+'±'+'NA'
                    else:
                        mpd_df.at[pop2, pop1] = str(round(pi, 5))+'±'+str(round(sd, 5))
            pop_num += 1

        return mpd_df

    # claculate pairwise Fst
    # Weir, B. S. and Hill, W. G. (2002) Estimating F-statistics. Annual Review of Genetics, 36, 721–750.
    def _fst(self, pops, population_data, pops_haplotype_data):

        # get allele frequency, sample size of each population
        pops_count, sample_sizes = self._calculate_allele_freq(pops, population_data, pops_haplotype_data)

        # calculate MSP and MSG to get theta statistic
        pops_freq = [np.array(i) / j for i, j in zip(pops_count, sample_sizes)]
        mean_freq = np.sum([np.array(i) for i in pops_count]) / np.sum(sample_sizes)


        MSP = np.sum([i*((j-mean_freq)**2) for i, j in zip(sample_sizes, pops_freq)], axis=0)

        MSG1 = float(1/np.sum([i-1 for i in sample_sizes]))
        MSG2 = np.sum([np.array(i)*(1-j) for i, j in zip(pops_count, pops_freq)], axis=0)
        MSG = MSG1 * MSG2

        nc = np.sum(sample_sizes) - np.sum(sample_sizes**2)/np.sum(sample_sizes)

        theta1 = np.nansum(MSP)-np.nansum(MSG)
        theta2 = np.nansum(MSP)+(nc-1)*np.nansum(MSG)
        if theta2 == 0:
            theta = 0
        else:
            theta = (theta1 / theta2)

        return theta

    # calculate pairwise Rst
    # Slatkin, M. (1995) A measure of population subdivision based on microsatellite allele frequencies. Genetics, 139, 457–462.
    def _rst(self, pops, population_data, distance_matrix):

        n = len(distance_matrix)
        pop_num = len(pops)
        expr = 2 / (pop_num*(pop_num-1))

        level_labels = population_data[0:, self.label_num]
        labels_num = []
        for i in Counter(level_labels).values():
            labels_num += [i]*i
        level_labels_code = [sorted(list(set(level_labels))).index(i) for i in level_labels]

        # calculate SW
        SW = 0
        bool1 = np.equal.outer(level_labels_code, level_labels_code)
        for i in range(pop_num):
            bool2 = np.tile(np.array(level_labels_code) == 1, (n, 1))
            bool_SW = bool1 & bool2 & bool2.T
            SW += np.sum(distance_matrix[bool_SW]) / np.sum(bool_SW)
        SW = SW / pop_num

        # calculate SB
        SB = 0
        bool3 = np.not_equal.outer(level_labels_code, level_labels_code)
        for i in range(pop_num-1):
            for j in range(i+1, pop_num):
                bool4 = np.tile(np.array(level_labels_code) == i, (n, 1)) | np.tile(np.array(level_labels_code) == j, (n, 1))
                bool_SB = bool3 & bool4 & bool4.T
                SB += np.sum(distance_matrix[bool_SB]) / np.sum(bool_SB)
        SB = SB * expr

        n_bar = np.mean(list(Counter(level_labels).values()))
        denom = np.sum(list(Counter(level_labels).values())) - 1
        S_bar = (SW*(n_bar-1)/denom) + (SB*(n_bar*(pop_num-1))/denom)
        res = (S_bar-SW) / S_bar

        return res

    # calculate hs  standard hs and ht of Gst
    def _gst(self, pops, population_data, pops_haplotype_data):

        # get allele frequency, sample size of each population
        pops_count, sample_sizes = self._calculate_allele_freq(pops, population_data, pops_haplotype_data)

        pops_freq = np.array([np.array(i) / j for i, j in zip(pops_count, sample_sizes)])

        hs = 1 - np.nansum(np.nansum((pops_freq**2), axis=0), axis=1)/len(pops)
        ht = 1 - np.nansum((np.nansum(pops_freq**2, axis=2)/len(pops))**2, axis=0)

        return hs, ht, sample_sizes

    # calculate Nei Gst
    # Nei M (1973) Analysis of Gene Diversity in Subdivided Populations. Proc. Nat. Acad. Sci., 70, 3321-3323.
    def _Nei_gst(self, pops, population_data, pops_haplotype_data):

        hs, ht, sample_sizes = self._gst(pops, population_data, pops_haplotype_data)
        Nei_gst = 1 - (np.mean(hs)/np.mean(ht))

        return Nei_gst

    # calculate Nei and Chesser Gst
    # Nei M, Chesser RK (1983) Estimation of fixation indices and gene diversity. Annals of Human Genetics, 47, 253-259.
    def _Nei_Chesser_gst(self, pops, population_data, pops_haplotype_data, return_flag=False):

        hs, ht, sample_sizes = self._gst(pops, population_data, pops_haplotype_data)

        nm = len(pops) / np.sum(1/sample_sizes)
        Hs = (nm/(nm-1)) * hs
        Ht = ht + Hs/(nm*len(pops))

        Nei_Chesser_gst = 1 - np.mean(Hs)/np.mean(Ht)

        if return_flag:
            return Hs, Nei_Chesser_gst
        else:
            return Nei_Chesser_gst

    # calculate Hedrick Gst
    # Hedrick P (2005) A standardized genetic differentiation measure. Evolution, 59, 1633-1638.
    def _Hedrick_gst(self, pops, population_data, pops_haplotype_data):

        Hs, Nei_Chesser_gst = self._Nei_Chesser_gst(pops, population_data, pops_haplotype_data, True)

        Hedrick_gst = Nei_Chesser_gst * (len(pops)-1+np.mean(Hs))/((len(pops)-1)*(1-np.mean(Hs)))

        return Hedrick_gst

    # get pairwise genetic distance of two populations
    def _get_gd_and_p(self, pops, gd_func):

        pops_haplotype_data = self._get_population_data(pops, self.population_array) # haplotype data of specific populations
        population_data = self.population_array[list(map(lambda x: x in pops, self.population_array[:, self.label_num]))] # population fata of specific populations

        # calculate genetic distance (Fst, Rst or Dst)
        if gd_func == self._rst:
            inds = list(range(len(pops_haplotype_data)))
            distance_matrix = np.zeros(shape=(len(inds), len(inds)))
            for ind1 in inds:
                seq1 = pops_haplotype_data[ind1]
                for ind2 in inds:
                    if ind1 == ind2:
                        distance = 0.0
                    else:
                        seq2 = pops_haplotype_data[ind2]
                        distance = self._pairwise_distance(seq1, seq2, 'sum_of_squared_diff')
                    distance_matrix[ind1, ind2] = distance
            gd = gd_func(pops, population_data, distance_matrix)
        else:
            gd = gd_func(pops, population_data, pops_haplotype_data)

        # permutation to get genetic distance
        permuted_index = list(range(0, len(population_data)))
        permuted_population_data = population_data.copy()
        permuted_gds = []
        for i in range(1000):
            random.shuffle(permuted_index)
            permuted_pops = population_data[permuted_index, self.label_num]
            permuted_population_data[:, self.label_num] = permuted_pops
            if gd_func == self._rst:
                permuted_gd = gd_func(pops, permuted_population_data, distance_matrix)
            else:
                permuted_gd = gd_func(pops, permuted_population_data, pops_haplotype_data)
            permuted_gds.append(permuted_gd)

        # calculate p-value
        p_value = np.sum(np.array(permuted_gds) >= gd) / (1000+1)
        p_value = round(p_value, 5)
        if p_value == 0:
            lowest_p_value = round(1/1001, 5)
            p_value = '<%s' % lowest_p_value

        return gd, p_value

    # get pairwise and global genetic distance of all populations
    def _calculate_gd(self, gd_type):

        # create empty gnetic diastance matrix
        gd_df = pd.DataFrame(index=self.pops, columns=self.pops)
        # function used to calculate genetic distances
        if gd_type == 'fst':
            gd_func = self._fst
        elif gd_type == 'rst':
            gd_func = self._rst
        elif gd_type == 'Ngst':
            gd_func = self._Nei_gst
        elif gd_type == 'NCgst':
            gd_func = self._Nei_Chesser_gst
        elif gd_type == 'Hgst':
            gd_func = self._Hedrick_gst

        # calculate pairwise genetic distances
        pop_num = 0
        for pop1 in self.pops:
            pop_num += 1
            for pop2 in self.pops[pop_num:]:
                pops = [pop1, pop2]
                # calculate pairwise genetic distance and p value
                p_gd, p_gd_p = self._get_gd_and_p(pops, gd_func)
                gd_df.at[pop1, pop2] = p_gd_p
                gd_df.at[pop2, pop1] = p_gd

        gd_df = gd_df.fillna('-')

        # calculate global genetic distances
        g_gd, g_gd_p = self._get_gd_and_p(self.pops, gd_func) # calculate global fst or rst and p value

        return gd_df, g_gd, g_gd_p

    # get df of AMOVA
    def _amova_df(self, amova_pop_data, N, total_level_num, levels):

        # pairwise: df between, df within, df total
        # total: df among, df within, df total
        df = ['']*(total_level_num+2)
        df[:total_level_num] = [len(set(amova_pop_data[:, i])) for i in levels[::-1]]
        df[total_level_num] = N
        df[1: total_level_num+1] = [df[i]-df[i-1] for i in range(1, total_level_num+1)] # df within
        df[0] -= 1 # df between/among
        df[-1] = N - 1 # df total
        df = np.array(df)

        return df

    # get SSD of AMOVA
    def _amova_SSD(self, amova_pop_data, distance_matrix, N, total_level_num, levels):

        SSD = ['']*(total_level_num+2) # SSD between/among, SSD within, SSD total
        SSD[-1] = np.sum(distance_matrix)/float(2*N) # SSD total
        # calculate SSD within
        for level_num, level_name in enumerate(levels[::-1]):
            level_labels = amova_pop_data[:, level_name].tolist()
            labels_num = []
            for i in Counter(level_labels).values():
                labels_num += [i]*i
            level_labels_code = [sorted(list(set(level_labels))).index(i) for i in level_labels]
            SSD[level_num+1] = np.sum((distance_matrix/(2*np.array(labels_num)))[np.equal.outer(level_labels_code, level_labels_code)])
        if total_level_num == 2: # Group in levels
            SSD[1] = SSD[1] - SSD[2]
        SSD[0] = SSD[-1] - sum(SSD[1:-1]) # SSD between/among

        return SSD

    # get coefficient of variance in AMOVA
    def _amova_coef(self, amova_pop_data, N, total_level_num):

        pop_num = len(set(amova_pop_data[:, 1]))
        pop_level_count = np.array(list(Counter(amova_pop_data[:, 1]).values()))
        pop_level_count_square = pop_level_count**2

        if total_level_num == 1:
            # calculate coefficient of variance a
            coef = (N-np.sum(pop_level_count_square)/N) / (pop_num-1)
        elif total_level_num == 2:
            coef = [''] * 3
            group_num = len(set(amova_pop_data[:, 2]))
            # count how many populations belong to a group
            # get each first group-pop of group
            new_array = [tuple(row) for row in amova_pop_data[:, 1:]]
            group_pop = np.unique(new_array, axis=0)[:, 2-1]
            group_count_in_group_pop = list(Counter(group_pop).values()) # the count of group
            # count the sum of population freq in each group
            pop_count_in_group = []
            pop_count_in_group_squre = []
            group_labels = sorted(list(set(amova_pop_data[:, 2])))
            for i in group_labels:
                group_data = amova_pop_data[amova_pop_data[:, 2]==i]
                group_count = np.array(list(Counter(group_data[:, 1]).values()))
                pop_count_in_group.append(np.sum(group_count))
                pop_count_in_group_squre.append(np.sum(group_count**2))
            ratio = np.sum(np.array(pop_count_in_group_squre) / np.array(pop_count_in_group))
            # calculate coefficient of variance
            coef[0] = (N-ratio) / np.sum(np.array(group_count_in_group_pop)-1) # n1
            coef[1] = (ratio-np.sum(pop_level_count_square)/N) / (group_num-1) # n2
            coef[2] = (N-np.sum(np.array(pop_count_in_group)**2/N)) / (group_num-1) # n3
        coef = np.round(np.array(coef), 5)

        return coef

    # get Fst or Rst based on AMOVA analysis
    # get AMOVA table
    def _amova(self, pops, type, global_flag=False):

        if global_flag:
            if self.label == 'Population':
                levels = [self.label_num]
            else:
                levels = list(range(1, np.size(self.population_array, 1)))
            amova_pop_data = self.population_array
        else:
            levels = [self.label_num]
            amova_pop_data = self.population_array[list(map(lambda x: x in pops, self.population_array[:, self.label_num]))]

        amova_inds = amova_pop_data[:, 0]
        amova_haplotype_data = self.haplotype_array[list(map(lambda x: x in amova_inds, self.haplotype_array_with_inds[:, 0]))]
        total_level_num = len(levels) # number of levels: population or group
        N = len(amova_pop_data) # sample size

        # get distance matrix of two populations
        distance_matrix = np.zeros(shape=(len(amova_inds), len(amova_inds)))
        for ind1 in range(len(amova_inds)):
            seq1 = amova_haplotype_data[ind1]
            for ind2 in range(len(amova_inds)):
                if ind1 == ind2:
                    distance = 0
                else:
                    seq2 = amova_haplotype_data[ind2]
                    distance = self._pairwise_distance(seq1, seq2, type)
                distance_matrix[ind1, ind2] = distance

        # calculate df
        df = self._amova_df(amova_pop_data, N, total_level_num, levels)

        # calculate SSD
        SSD = self._amova_SSD(amova_pop_data, distance_matrix, N, total_level_num, levels)

        # calculate MSD
        MSD = SSD / df

        # # calculate coefficient and variance
        coef = self._amova_coef(amova_pop_data, N, total_level_num)
        sigma = ['']*(total_level_num+1)
        if total_level_num == 1:
            sigma[1] = MSD[1] # sigma b
            sigma[0] = (MSD[0]-MSD[1])/coef # sigma a
            statistics = [sigma[0] / (sigma[0]+sigma[1])] # phi_ST
        elif total_level_num == 2:
            sigma[2] = MSD[2] # sigma c
            sigma[1] = (MSD[1]-sigma[2]) / coef[0] # sigma b
            sigma[0] = (MSD[0]-MSD[2]-coef[1]*sigma[1]) / coef[2] # sigma a
            statistics = ['']*3
            statistics[0] = (sigma[0]) / (sigma[0]+sigma[1]+sigma[2]) # phi_CT
            statistics[1] = sigma[1] / (sigma[0]+sigma[1]) # phi_SC
            statistics[2] = (sigma[0]+sigma[1]) / (sigma[0]+sigma[1]+sigma[2]) # phi_ST
        sigma = np.round(np.array(sigma), 5)
        statistics = np.round(np.array(statistics), 5)

        # permutaion to get p value
        p_value = ['']*(total_level_num)
        r_sigma =  np.zeros(shape=(1000, total_level_num)) # sigma number = level number
        if total_level_num == 1:
            for i in range(1000):
                perm_order = np.random.permutation(np.arange(N)) # random permutation order
                r_distance_matrix = distance_matrix[np.ix_(perm_order, perm_order)] # permuted distance matrix
                r_SSD = self._amova_SSD(amova_pop_data, r_distance_matrix, N, total_level_num, levels)
                r_MSD = r_SSD / df
                r_sigma[i, 0] = (r_MSD[0]-r_MSD[1])/coef
        elif total_level_num == 2:
            group_label_count = list(Counter(amova_pop_data[:, 2]).values())
            end = 0
            label_num = []
            for i, j in enumerate(group_label_count):
                end += j
                start = sum(group_label_count[:i])
                label_num.append([start, end])
            for i in range(1000):
                perm_order = np.array([])
                for num in label_num:
                    perm_order = np.append(perm_order, np.random.permutation(np.arange(num[0], num[1])))
                perm_order = perm_order.astype('int64')
                r_distance_matrix = distance_matrix[np.ix_(perm_order, perm_order)]
                r_SSD = self._amova_SSD(amova_pop_data, r_distance_matrix, N, total_level_num, levels)
                r_MSD = r_SSD / df
                r_sigma[i, 1] = (r_MSD[1]-r_MSD[2]) / coef[0]
            pop_label_count = list(Counter(amova_pop_data[:, 1]).values())
            new_array = [tuple(row) for row in amova_pop_data[:, levels]]
            pop_group = np.unique(new_array, axis=0)[:, 2-1]
            r_amova_pop_data = amova_pop_data.copy()
            for i in range(1000):
                random.shuffle(pop_group)
                r_amova_pop_data[:, 2] = list(chain.from_iterable(repeat(group, count) for group, count in zip(pop_group, pop_label_count)))
                r_SSD = self._amova_SSD(r_amova_pop_data, r_distance_matrix, N, total_level_num, levels)
                r_df = self._amova_df(r_amova_pop_data, N, total_level_num, levels)
                r_MSD = r_SSD / r_df
                r_coef = self._amova_coef(r_amova_pop_data, N, total_level_num)
                r_sigma[i, 0] = (r_MSD[0]-r_MSD[2]-r_coef[1]*((r_MSD[1]-r_MSD[2])/r_coef[0])) / r_coef[2]
        for i in range(total_level_num):
            p_value[i] = np.round(np.sum(r_sigma[:, i] >= sigma[i]) / (1000+1), 5)

        if 0 in p_value:
            lowest_p_value = round(1/1001, 5)
            p_value = ['<%s' % lowest_p_value if i == 0 else i for i in p_value]

        if global_flag:
            return SSD, df, MSD, sigma, p_value, statistics, coef
        else:
            return statistics[0], p_value[0]

    # calculate pairwise Fst or Rst of populations
    def _calculate_amova(self, stat_type):

        if stat_type == 'fst':
            type = 'num_of_diff_alleles'
        elif stat_type == 'rst':
            type = 'sum_of_squared_diff'

        stat_df = pd.DataFrame(index=self.pops, columns=self.pops)
        pop_num = 0
        for pop1 in self.pops:
            pop_num += 1
            for pop2 in self.pops[pop_num:]:
                pops = [pop1, pop2]
                stat, p = self._amova(pops, type)
                stat_df.at[pop1, pop2] = p
                stat_df.at[pop2, pop1] = stat

        stat_df = stat_df.fillna('-')

        # Global AMOVA
        amova_items = self._amova(self.pops, type, True) # SSD, Df, MSD, sigma, statistics, p_value
        # AMOVA table
        if self.label == 'Population':
            amova_df_index = ['Among_populations', 'Within_populations', 'Total']
            variance_df_index = ['a', 'b']
            phi_df_columns = ['Phi_ST']
            coef_df_columns = ['n1']
        elif self.label == 'Group':
            amova_df_index = ['Among_groups', 'Among_populations/Within_Groups', 'Within_populations', 'Total']
            variance_df_index = ['a', 'b', 'c']
            phi_df_columns = ['Phi_CT', 'Phi_SC', 'Phi_ST']
            coef_df_columns = ['n1', 'n2', 'n3']
        amova_df = pd.DataFrame(index=amova_df_index, columns=['SSD', 'Df', 'MSD'])
        variance_df = pd.DataFrame(index=variance_df_index, columns=['Variance', 'p_value'])
        phi_df = pd.DataFrame(index=['Phi_statistic'], columns=phi_df_columns)
        coef_df = pd.DataFrame(index=['Variance_coefficients'], columns=coef_df_columns)
        for i, j in zip(amova_items[:-2], ['SSD', 'Df', 'MSD', 'Variance', 'p_value']):
            if j in amova_df.columns:
                tmp_df = amova_df
            else:
                tmp_df = variance_df
            if len(i) != tmp_df.index.size:
                diff = tmp_df.index.size - len(i)
                tmp_df[j] = np.append(i, np.array(['']*diff))
            else:
                tmp_df[j] = i

        phi_df.loc['Phi_statistic'] = amova_items[-2]
        coef_df.loc['Variance_coefficients'] = amova_items[-1]

        return stat_df, amova_df, variance_df, phi_df, coef_df

    # output calaulation result
    def calculate_and_output(self, arguments, path):

        if arguments.hd:
            hd_df = self._calculate_haplotype_diversity()
            hd_df.to_csv(path+'.HD.%s.txt' % self.label, sep='\t')
            self.logger.info('[Y-LineageTracker] Calculate haplotype diversity finished')
        if arguments.mpd:
            mpd_df = self._calculate_mean_pairwise_differences()
            mpd_df.to_csv(path+'.MPD.%s.txt' % self.label, sep='\t')
            self.logger.info('[Y-LineageTracker] Calculate mean pairwise differences on %s level finished' % self.label)
        if arguments.gd:
            self.logger.info('[Y-LineageTracker] Calculating pairwise genetic distance on %s level and p-value from permutation……' % self.label)
            gd_df, global_gd, global_gd_p = self._calculate_gd(arguments.gd)
            gd_out = open(path+'.GD.%s.%s.txt' % (arguments.gd, self.label), 'w')
            gd_out.write(gd_df.to_string()+'\n\n')
            gd_out.write('Global %s: %f\n' % (arguments.gd, global_gd))
            gd_out.write('p-value of Global %s: %s' % (arguments.gd, global_gd_p))
            gd_out.close()
            self.logger.info('[Y-LineageTracker] Calculation of pairwise genetic distance and p-value on %s level finished' % self.label)
        if arguments.amova:
            self.logger.info('[Y-LineageTracker] Calculating pairwise %s based on AMOVA of %s level……' % (arguments.amova, self.label))
            stat_df, amova_df, variance_df, phi_df, coef_df = self._calculate_amova(arguments.amova)
            stat_df.to_csv(path+'.AMOVA.%s.%s.txt' % (arguments.amova, self.label), sep='\t')
            self.logger.info('[Y-LineageTracker] Calculating pairwise %s based on AMOVA of %s level finished' % (arguments.amova, self.label))
            amova_df_out = open(path+'.AMOVA.%s.txt' % (self.label), 'w')
            amova_df_out.write(amova_df.to_string()+'\n\n'+variance_df.to_string()+'\n\n'+phi_df.to_string()+'\n\n'+coef_df.to_string())
            amova_df_out.close()
            self.logger.info('[Y-LineageTracker] AMOVA on %s level finished' % self.label)


# predict Y-haplogroup from Y-STRs
def predict_haplogroup(haplotype_data, path):

    logger = logging.getLogger()
    skip_letters = config.get('Statistics', 'SkipLetters').split(',')
    header = haplotype_data.columns.tolist()
    haplotype_array = np.array(haplotype_data)

    # get Y-STR prediction panel
    from FilesIO import CommonData
    common_data = CommonData()
    predict_panel = common_data.read_str_prediction_panel()
    array_columns = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'L', 'M', 'N', 'O', 'Q', 'R', 'S', 'T']

    to_be_imputed = 'Haplogroup'
    predict_panel[to_be_imputed] = predict_panel[to_be_imputed].map(lambda x: x[0])

    code, code2 = 0, 0
    str_list, str_list2 = [], []
    sample_list = []
    array_list = []
    str_panel = {}
    used = {}
    str_info = {}
    # get samples and their Y-STRs
    for n, sample in enumerate(haplotype_data.index):
        STR = {}
        for m, STR_name in enumerate(header):
            STR_num = haplotype_array[n, m]
            if STR_num not in skip_letters:
                try:
                    STR[STR_name] = int(STR_num)
                except ValueError:
                    logger.wrning('[Y-LineageTracker] [Warning] The STR input of %s does not meet the requirements' % sample)

        # prediction panel for each ind
        STRs = list(STR.keys())
        if STRs in str_list:
            code = str_list.index(STRs)
            used_STRs = used[code]
            ind_predict_panel = str_panel[code]
        else:
            na_count = predict_panel[STRs].isnull().sum(axis=0)
            used_STRs = na_count[na_count<1000].index.tolist()
            if len(used_STRs) < 6:
                ind_predict_panel = None
            else:
                ind_predict_panel = predict_panel[[to_be_imputed]+used_STRs]
            used[code] = used_STRs
            str_panel[code] = ind_predict_panel
            str_list.append(STRs)
            code += 1
        if len(used_STRs) < 6:
            logger.warning('[Y-LineageTracker] [Warning] The number of Y-STR loci is too small for sample %s' % sample)
            continue

        freq_list = []

        # calculate frequency
        ind_predict_hap = np.array(ind_predict_panel[to_be_imputed])
        STR_loci = ind_predict_panel.drop(to_be_imputed, axis=1).columns.tolist()
        ind_predict_panel = np.array(ind_predict_panel.drop(to_be_imputed, axis=1))

        ind_ref_panel = np.empty(shape=(0, np.size(ind_predict_panel, 1)))
        ind_ref_hap = []
        for i in sorted(list(set(ind_predict_hap))):
            sub = ind_predict_panel[ind_predict_hap==i]
            if len(sub) > 500:
                idx = np.random.randint(0, len(sub), 500)
                sub = sub[idx]
            ind_ref_panel = np.concatenate([ind_ref_panel, sub])
            ind_ref_hap.extend([i]*len(sub))
        ind_ref_hap = np.array(ind_ref_hap)
        to_be_imputed_freq = np.array(list(zip(*sorted(Counter(ind_ref_hap).items())))[1]).astype(float)

        # sample size control
        panel = {}
        array_location = {}
        for i, locus in enumerate(STR_loci):
            array_index = sorted(np.unique(ind_ref_panel[:, i]))
            array = np.zeros(shape=(len(array_index), len(array_columns)))
            for j, hap in enumerate(array_columns):
                col_data = ind_ref_panel[list(map(lambda x: x == hap, ind_ref_hap))][:, i]
                col_freq = Counter(col_data)
                col = [col_freq[i]/len(col_data) if i in col_freq.keys() else 0 for i in array_index]
                array[:, j] = col
            panel[locus] = array
            array_location[locus] = array_index

        # calculate the probability of each main haplogroup
        Pr_up = to_be_imputed_freq
        for i in STR.keys():
            if i in used_STRs:
                table_index = array_location[i].index(STR[i])
                Pr_up *= panel[i][table_index]
        D = sum(Pr_up)
        Pr = np.round(Pr_up/D, 4)
        array_list.append(Pr)
        sample_list.append(sample)

        percent = float(n + 1) * 100 / float(len(haplotype_data))
        sys.stdout.write('[Y-LineageTracker] Predicting haplogroups from Y-STR data…… %.2f' % percent)
        sys.stdout.write('%\r')
        sys.stdout.flush()

    # output result
    predict_df = pd.DataFrame(index=sample_list,
                              columns=sorted(list(set(predict_panel['Haplogroup'].map(lambda x: x[0])))),
                              data=np.array(array_list))
    predict_df.to_csv(path+'.Prediction.txt', sep='\t')

    print()
    logger.info('[Y-LineageTracker] Haplogroup prediction finished')


def main():

    start = time.perf_counter()
    arguments = stat_parser()

    from FilesIO import check_seq_matrix_input, check_hap_input, set_haplotype_index, check_population, check_overlap, check_pop_header, time_count
    # check data input
    input, format = check_seq_matrix_input(arguments.seq, arguments.format, arguments.matrix)
    haplotype_data = check_hap_input(input, 'haplotype', format)

    # check commands
    arguments.amova = check_amova(arguments.amova, sys.argv)
    check_commands(arguments)

    from FilesIO import get_out_path
    # get path of output files
    path = get_out_path(arguments.output, 'Stat')
    log_file = path + '.StatLog.log'

    set_log(arguments, log_file)

    # check population data
    population_data = check_population(arguments.population)
    population_data = check_pop_header(population_data)
    haplotype_data, population_data = check_overlap(haplotype_data, population_data)
    haplotype_data = set_haplotype_index(haplotype_data, format)

    # calculate statistics
    label = 'Population'
    stat = StatisticsAnalysis(label, haplotype_data, population_data)
    stat.calculate_and_output(arguments, path)
    if 'Group' in population_data.columns:
        label = 'Group'
        stat = StatisticsAnalysis(label, haplotype_data, population_data)
        stat.calculate_and_output(arguments, path)

    # predict haplogroup from Y-STR data
    if arguments.predict and arguments.matrix:
        predict_haplogroup(haplotype_data, path)

    time_count(start)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#-*- coding:utf-8 -*-

import os
import sys

'''
information about version and module
'''

__VERSION__ = '1.3.0'
__MODULE__ = 'Y-LineageTracker'

current_dir = os.path.split(os.path.realpath(__file__))[0]
sys.path.append(current_dir)
sys.path.append('..')



import os
os.environ['QT_QPA_PLATFORM']='offscreen'
import gc
import sys
import time
import logging
import argparse
import matplotlib.colors
import numpy as np
import pandas as pd

from Bio import Phylo
from itertools import filterfalse
from ete3 import Tree
from GetConfig import getConfig


'''
Description:
This module is used for phylogeny analysis.
It is recommended to input both haplogroup file and sequence alignment file to construct a bifurcating tree.
If only the haplogroup file is given, the phylogenetic tree may contain polytomies (a node with more than two children).
For polytomies, if sequence alignment is provided, the program will apply alignment method to construct a sub-tree for nodes.
'''


config = getConfig()


def phylo_parser():

    parser = argparse.ArgumentParser('phylo', description='(c) Y-LineageTracker: Phylogeny analysis')
    # function used for phylogeny analysis
    parser.add_argument('phylo',
                        help='Phylogenetic analysis for NRY haplogroups.')
    # required, haplogroup file
    parser.add_argument('--hg',
                        required=True,
                        type=str,
                        action='store',
                        help='hg: A file containing sample ID and haplogroup of each individual.')
    # optional, sequence alignment file
    parser.add_argument('--seq',
                        required=False,
                        type=str,
                        action='store',
                        help='seq: Sequence alignment file of input samples.')
    # optional, format of sequence alignment
    parser.add_argument('--seq-format',
                        required=False,
                        type=str,
                        dest='format',
                        choices=['fasta', 'phylip', 'nexus', 'meg', 'vcf'],
                        help='seq-format: The file format of sequence alignment file. The default is fasta. ')
    # optional, population file
    parser.add_argument('-p', '--population',
                        required=False,
                        type=str,
                        action='store',
                        help='population: A file containing sample ID and population information of each individual')
    # optional, method of calculating distance from sequence alignment
    parser.add_argument('--align-method',
                        required=False,
                        type=str,
                        dest='align',
                        action='store',
                        default='mp',
                        choices=['ibs', 'upgma', 'mp'],
                        help='align-method: The method of constructing the bifurcating tree from multiple sequences. The default is mp.')
    # optional, the output format in text of tree
    parser.add_argument('--tree-format',
                        required=False,
                        type=str,
                        dest='tree',
                        action='store',
                        default='newick',
                        choices=['newick', 'nexus', 'phyloxml', 'nexml'],
                        help='--tree-format: The file format of output phylogenetic tree. The default is newick. ')
    # optional, the layout of tree in figure
    parser.add_argument('--layout',
                        required=False,
                        type=str,
                        action='store',
                        default='rectangular',
                        choices=['rectangular', 'circular'],
                        help='layout: The layout of the phylogenetic tree figure. The default is rectangular.')
    # optional, the prefix of output
    parser.add_argument('-o', '--output',
                        required=False,
                        type=str,
                        action='store',
                        help='output: The prefix of output files.')


    args = parser.parse_args()

    return args


# print program information and write to log file
def set_log(args_log, log_file):

    logger = logging.getLogger()
    logger.setLevel(level=logging.INFO)

    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] - [%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    logger.addHandler(handler)
    logger.addHandler(console)

    log_info = ['[Y-LineageTracker] [Phylo]',
                '[Y-LineageTracker] Run Date: ' + time.asctime(time.localtime(time.time())),
                '[Y-LineageTracker] Haplogroup File: %s' % args_log.hg,
                '[Y-LineageTracker] Seq File: %s' % args_log.seq,
                '[Y-LineageTracker] Seq Format: %s' % args_log.format,
                '[Y-LineageTracker] Population File: %s' % args_log.population,
                '[Y-LineageTracker] Tree Layout: %s' % args_log.layout,
                '[Y-LineageTracker] Output Tree Format: %s' % args_log.tree]

    print('\n')
    for i in log_info:
        logger.info(i)


# names of output files
def check_phylo_output(tree_format, output):

    from FilesIO import get_out_path
    path = get_out_path(output, 'Phylo')
    fig_file = path + '.phylo'
    if 'xml' in tree_format:
        tree_file = path + '.tree.xml'
    else:
        if tree_format == 'newick':
            tree_file = path + '.tree.nwk'
        elif tree_format == 'nexus':
            tree_file = path + '.tree.nex'
    log_file = path + '.PhyloLog.log'
    output_set = [fig_file, tree_file, log_file]

    return output_set


class PhyloHaplogroup(object):
    '''
    This class is used to perform main phylogeny analysis
    It can:
    1. construct a haplogroup tree form haplogroup file
    2. construct sub-trees for internal polytomies
    3. output tree with a figure
    Population file is recommended for tree plot to construct a bifurcating tree
    Output tree file in newick format can be used as one of input file in time estimation
    '''

    def __init__(self, hap_data, output_set):

        self.logger = logging.getLogger()
        hap_data['Haplogroup'] = hap_data['Haplogroup'].map(lambda x: x.strip('~'))
        self.hap_data = hap_data
        self.output_set = output_set
        self.common_trunk = config.get('HaplogroupTree', 'CommonTrunk').split(',')

    # read sequence alignment file
    def read_seq_file(self, seq_file, format):

        from ProcessData import ConvertData
        from Bio import AlignIO

        # get mutiple alignment data
        if format == 'vcf':
            convert = ConvertData.ConvertData(seq_file)
            msa = convert.data_convert('vcf', 'msa')
        elif format == 'phylip':
            msa = AlignIO.read(open(seq_file), 'phylip-relaxed')
        else:
            msa = AlignIO.read(open(seq_file), format)

        # get ids of msa
        ids = [i.id for i in msa]
        seqs = [i.seq for i in msa]

        return ids, seqs

    # get base tree from haplogroup file
    def get_base_tree(self, used_samples=None):

        self.logger.info('[Y-LineageTracker] Constructing phylogenetic tree...')

        # preprocess data
        hap_df = self.hap_data.sort_values(by='Haplogroup')
        if used_samples:
            hap_df = hap_df[hap_df['SampleID'].map(lambda x: x in used_samples)]
        hap_phylo = hap_df.drop_duplicates(subset='Haplogroup', keep='first', inplace=False)
        hap_inds = hap_df.set_index(['Haplogroup'], drop=True)

        # prepare variables
        new_tree_info = []
        filter_trunk = []
        from FilesIO import CommonData
        common_data = CommonData()
        tree_info = common_data.read_tree_info()

        # traverse tree to build a pylogenetic tree based on haplogroup
        for i in tree_info:
            hap_name = i.split('\t')[-1]
            if hap_name in self.common_trunk:
                new_tree_info.append(i)
                continue
            level = i.count('\t')
            if len(list(filterfalse(lambda x: not x.startswith(hap_name), hap_phylo['Haplogroup']))) == 0:
                new_tree_info.append(i)
                filter_trunk.append(hap_name)
                continue

            hap_point_list = list(filterfalse(lambda x: x == hap_name or not x.startswith(hap_name), hap_phylo['Haplogroup']))
            hap_point_list = [i for i in hap_point_list if i not in self.common_trunk]
            hap_dict = {}
            hap_class = []
            if len(hap_point_list) != 0 and len(set(hap_point_list)&set([i.split('\t')[-1] for i in tree_info])) == 0:
                num = 0
                hap_count = len((hap_point_list))+1
                iter_time = len(''.join(hap_point_list))
                if len(hap_point_list) == 1:
                    new_tree_info.append(i)
                    new_tree_info.append('-\t' * (level+1) + hap_point_list[0])
                    continue
                for j in range(iter_time):
                    if num+1 < len(hap_point_list):
                        length = len(hap_name)
                        if len(hap_point_list[num]) == length:
                            num += 1
                            hap_name = i.split('\t')[-1]
                            hap_count = len(hap_dict[hap_name])
                            continue
                        if hap_name in hap_class:
                            hap_count = len(hap_dict[hap_name])
                            hap_name = hap_point_list[num][:length+1]
                            continue
                        if len(list(filterfalse(lambda x: not x.startswith(hap_name), hap_point_list))) < hap_count and len(list(filterfalse(lambda x: not x.startswith(hap_name), hap_point_list))) > 1:
                            hap_dict[hap_name] = [i for i in hap_point_list if i.startswith(hap_name)]
                            hap_class.append(hap_name)
                            hap_count = len(hap_dict[hap_name])
                            hap_name = hap_point_list[num][:length+1]
                        else:
                            hap_name = hap_point_list[num][:length+1]
            if hap_dict:
                level_list = [i]
                key_list = sorted(list(hap_dict.keys()))
                key_list.reverse()
                ele_list = []
                for key in key_list:
                    one_key_list = hap_dict[key]
                    ele_list.extend(one_key_list)
                ele_list.extend(key_list)
                ele_list = sorted(list(set(ele_list)))
                level_count = 0
                level_hap = i.split('\t')[-1]
                level_dict = {}
                level_dict[i.split('\t')[-1]] = 0
                for ele_num, ele in enumerate(ele_list[1:]):
                    if ele.startswith(level_hap):
                        level_count += 1
                        level_dict[ele] = level_count
                    else:
                        for ele_back in reversed(ele_list[:ele_num]):
                            if ele.startswith(ele_back):
                                level_count = level_dict[ele_back]+1
                                level_dict[ele] = level_count
                                break
                    level_list.append('-\t' * (level+level_count) + ele)
                    level_hap = ele
                level_list = sorted(level_list, key=lambda level_list: level_list.split('-\t')[-1])
                if i == level_list[0]:
                    new_tree_info.extend(level_list)
            else:
                new_tree_info.append(i)

        # remove tree trunk not for tree construction
        intersection = [i for i in new_tree_info if i.split('\t')[-1] not in filter_trunk]
        for i in reversed(intersection):
            if i.split('\t')[-1] in self.common_trunk:
                common_level = i.count('\t')
                if intersection.index(i) == len(intersection)-1:
                    filter_trunk.append(i.split('\t')[-1])
                elif intersection[intersection.index(i)+1].count('\t') != common_level+1:
                    filter_trunk.append(i.split('\t')[-1])
                    if i.split('\t')[-1] == 'P1':
                        filter_trunk.append('P')
                    elif i.split('\t')[-1] == 'NO1':
                        filter_trunk.append('NO')
                elif all([hap.split('\t')[-1] in self.common_trunk or len(hap.split('\t')[-1]) == 1 for hap in intersection[intersection.index(i):]]):
                    filter_trunk.extend([hap.split('\t')[-1] for hap in intersection[intersection.index(i):]])

        # remove duplicated branch in tree construction
        rm_list = []
        iter_num = 0
        for i in new_tree_info:
            node_hap = i.split('\t')[-1]
            node_level = i.count('\t')
            if node_hap in self.common_trunk or len(node_hap) == 1:
                continue
            if new_tree_info.index(i) == len(new_tree_info)-1:
                continue
            if node_hap in hap_df['Haplogroup'].tolist():
                continue
            if new_tree_info[new_tree_info.index(i)+1].split('\t')[-1].startswith(node_hap):
                branch_hap = new_tree_info[new_tree_info.index(i)+1].split('\t')[-1]
                node_base = [i for i in new_tree_info if i.split('\t')[-1].startswith(node_hap) and i.split('\t')[-1]!= node_hap]
                branch_base = [i for i in new_tree_info if i.split('\t')[-1].startswith(branch_hap)]
                if node_base == branch_base:
                    if iter_num == 0:
                        new_tree_info_irredundant = [(i.count('\t')-1)*'-\t' + i.split('\t')[-1] if i.split('\t')[-1].startswith(node_hap) else i for i in new_tree_info]
                    else:
                        new_tree_info_irredundant = [(i.count('\t')-1)*'-\t' + i.split('\t')[-1] if i.split('\t')[-1].startswith(node_hap) else i for i in new_tree_info_irredundant]
                    iter_num += 1
                    rm_list.append(node_hap)
        new_tree_info = [i for i in new_tree_info_irredundant if i.split('\t')[-1] not in rm_list]

        # add SampleID to tree
        ind_tree_info = []
        min_level_num = config.getint('PlotParameters', 'MinLevelNum')
        for i in new_tree_info:
            ind_hap = i.split('-\t')[-1]
            level_num = i.count('-\t')
            if ind_hap in hap_phylo['Haplogroup'].tolist():
                ind_tree_info.append(i)
                if isinstance(hap_inds.at[ind_hap, 'SampleID'], str):
                    ind_tree_info.append((level_num+1) * '-\t' + hap_inds.at[ind_hap, 'SampleID'])
                else:
                    ind_list = hap_inds.at[ind_hap, 'SampleID'].tolist()
                    ind_tree_info = ind_tree_info + [(level_num+1) * '-\t' + i for i in ind_list]
                if level_num < min_level_num:
                    min_level_num = level_num
            elif ind_hap not in filter_trunk:
                ind_tree_info.append(i)
        max_length = max([len(i.split('-\t')[-1]) for i in new_tree_info])

        # tree in newick format for visulization
        samples = hap_df['SampleID'].tolist()
        base, trunk_length = self._write_nwk(ind_tree_info, hap_df['SampleID'].tolist())
        base = base[1:].replace(');', ';') # root

        # remove reduandant nodes
        base_tree = Tree(base, format=8)
        for node in base_tree.traverse('preorder'):
            if (not node.is_leaf()) and len(node.children) == 1:
                child_name = node.children[0].name
                if child_name in self.common_trunk or len(child_name) == 1:
                    node.delete()

        base = base_tree.write(format=8)
        base = base.replace('NoName', '')

        return hap_df, base, trunk_length, max_length, samples

    # output tree in specific format
    def output_tree(self, base, tree_format):

        if base.endswith(');'):
            base = base[1:].replace(');', ';') # root

        # output tree in specific format
        output_file = self.output_set[1]
        if tree_format == 'newick':
            output_tree = open(output_file, 'w')
            output_tree.write(base)
            output_tree.close()
        else:
            from io import StringIO
            tmp_tree = StringIO(base)
            nwk_tree = Phylo.parse(tmp_tree, 'newick')
            Phylo.write(nwk_tree, output_file, tree_format)

    # apply sequence alignment method to construct a bifurcating tree
    def phylo_internal(self, ids, seqs, base, samples, align):

        # change the internal node of tree
        base_tree = Tree(base, format=8)
        for node in base_tree.iter_descendants('postorder'):
            if len(node.children) > 2 and node.name != '':
                node_name = node.name
                tmp_node = Tree('(tmp);', format=8)
                children = node.children
                children_haps_num = len([i for i in children if i.name not in samples])
                children_sample_num = len([i for i in children if i.name in samples])
                if children_haps_num == 0: # all children samples
                    children_samples = [i for i in children if i.name in samples]
                    self._add_single_tree_from_tmp(node, tmp_node, children_samples, node_name, ids, seqs, align, True)
                elif children_haps_num == 1:
                    if children_sample_num == 2:
                        children_samples_name = [i.name for i in children if i.name in samples]
                        children_haps_name = [i.name for i in children if i.name not in samples]
                        polynode = base_tree.get_common_ancestor(children_haps_name[0], children_samples_name[0])
                        polynode.resolve_polytomy(recursive=False)
                    else:
                        children_samples = [i for i in children if i.name in samples]
                        self._add_single_tree_from_tmp(node, tmp_node, children_samples, node_name, ids, seqs, align)
                elif children_haps_num > 1:
                    hap_sample_dict = {}
                    if children_sample_num <= 1: # all sub-haplogroups
                        children_haps = [i for i in children if i.name not in samples]
                        self._add_multi_tree_from_tmp(node, tmp_node, children_haps, node_name, ids, seqs, samples, align)
                    else:
                        children_samples = [i for i in children if i.name in samples]
                        self._add_single_tree_from_tmp(node, tmp_node, children_samples, node_name, ids, seqs, align)
                        children_haps = [i for i in children if i.name not in samples]
                        tmp_node2 = Tree('(tmp2);', format=8)
                        self._add_multi_tree_from_tmp(node, tmp_node2, children_haps, node_name, ids, seqs, samples, align)
            gc.collect()

        # remove nodes without children
        for node in base_tree.traverse('preorder'):
            if (not node.is_leaf()) and len(node.children) == 1:
                child_name = node.children[0].name
                if child_name not in samples:
                    node.delete()

        base = base_tree.write(format=8)
        base = base.replace('NoName', '')

        return base

    def _add_single_tree_from_tmp(self, node, tmp_node, children_samples, node_name, ids, seqs, align, replace=False):

        node.add_child(tmp_node)
        children_samples_name = [i.name for i in children_samples]
        # delete parallel samples
        for i in children_samples:
            i.delete()
        # concat sequence_distance_tree to haplogroup node
        id_idx = [ids.index(i) for i in children_samples_name]
        children_seqs = [seqs[i] for i in id_idx]
        children_tree = self._get_sequence_distance_tree(children_seqs, children_samples_name, align)
        if replace:
            children_tree = children_tree.replace(';', node_name+';')
        node.search_nodes(name='tmp')[0].add_child(Tree(children_tree, format=8))
        node.search_nodes(name='tmp')[0].delete()

    def _add_multi_tree_from_tmp(self, node, tmp_node, children_haps, node_name, ids, seqs, samples, align, replace=False):

        node.add_child(tmp_node)
        tmp_name = [i.name for i in tmp_node][0]
        hap_sample_dict = {}
        for hap in children_haps:
            hap_children_sample = list(filter(lambda x: x != '' and x in samples,
                                              [i.name for i in hap.traverse()]))[0]
            hap_sample_dict[hap_children_sample] = hap.name
        hap_children_samples = list(hap_sample_dict.keys())
        id_idx = [ids.index(i) for i in hap_children_samples]
        hap_children_seqs = [seqs[i] for i in id_idx]
        hap_children_tree = self._get_sequence_distance_tree(hap_children_seqs, hap_children_samples, align)
        for hap_children_sample, hap_name in hap_sample_dict.items():
            hap_children_tree = hap_children_tree.replace(hap_children_sample, hap_name+tmp_name)
        tmp_tree = Tree(hap_children_tree, format=8)
        for hap in children_haps:
            node_to_add = Tree(hap.write(format=8), format=8)
            tmp_tree.search_nodes(name=hap.name+tmp_name)[0].add_child(node_to_add)
            tmp_tree.search_nodes(name=hap.name+tmp_name)[0].delete()
            hap.detach()
        node.search_nodes(name=tmp_name)[0].add_child(tmp_tree)
        node.search_nodes(name=tmp_name)[0].delete()

    # get sub-tree from multiple sequence alignment
    def _get_sequence_distance_tree(self, seqs, ids, align):

        new_seqs = []
        # prune sequences
        for i in range(len(seqs[0])):
            all_al = [s[i] for s in seqs]
            al = list(set(all_al))
            N_count = 0
            if 'N' in al:
                al.remove('N')
                N_count += all_al.count('N')
            if '-' in al:
                al.remove('-')
                N_count += all_al.count('-')
            N_rate = N_count / len(all_al)
            if len(al) != 1 and N_rate < 0.05:
                new_seqs.append(all_al)

        new_seqs = np.array(new_seqs).T.tolist()

        # IBS method
        if align == 'ibs':
            from scipy.cluster import hierarchy
            from scipy.spatial.distance import squareform

            pairwise_df = pd.DataFrame(index=ids, columns=ids)

            # get pairwise IBS
            ind_num = 0
            for ind1 in ids:
                for ind2 in ids[ind_num:]:
                    if ind1 == ind2:
                        pairwise_df.at[ind1, ind2] = 1
                    else:
                        ibs = self._calculate_ibs(seqs, ids, ind1, ind2)
                        pairwise_df.at[ind1, ind2] = ibs
                        pairwise_df.at[ind2, ind1] = ibs
                ind_num += 1
            pairwise_df = 1 - pairwise_df

            # convert distance matrix to hierarchy
            dm = hierarchy.linkage(squareform(np.array(pairwise_df)), 'average')
            dm_tree = hierarchy.to_tree(dm, False)
            children_tree = self._hier_to_newick(dm_tree, '', pairwise_df.index.tolist())
        else:
            import tempfile
            from Bio import Phylo
            from Bio.Seq import Seq
            from Bio.SeqRecord import SeqRecord
            from Bio.Align import MultipleSeqAlignment
            from Bio.Phylo.TreeConstruction import DistanceCalculator, DistanceTreeConstructor

            new_seqs = [''.join(i) for i in new_seqs]
            seq_list = [SeqRecord(Seq(j), id=i, name=i) for i, j in zip(ids, new_seqs)]
            aln = MultipleSeqAlignment(seq_list)
            calculator = DistanceCalculator('identity')

            # UPGMA method
            if align == 'upgma':

                constructor = DistanceTreeConstructor(calculator, 'upgma')

                tree = constructor.build_tree(aln)
                tree_tmp_file = tempfile.NamedTemporaryFile(mode='w')
                Phylo.write(tree, tree_tmp_file.name, 'newick')
                ete_tree = Tree(open(tree_tmp_file.name).read(), format=1)
                children_tree = ete_tree.write(format=9)

            # MP (maximum parsimony) method
            elif align == 'mp':

                from Bio.Phylo.TreeConstruction import ParsimonyScorer, NNITreeSearcher, ParsimonyTreeConstructor

                dm = calculator.get_distance(aln)
                constructor = DistanceTreeConstructor(calculator, 'nj')
                tree = constructor.build_tree(aln)
                scorer = ParsimonyScorer()
                searcher = NNITreeSearcher(scorer)
                constructor = ParsimonyTreeConstructor(searcher, tree)
                pars_tree = constructor.build_tree(aln)

                tree_tmp_file = tempfile.NamedTemporaryFile(mode='w')
                Phylo.write(pars_tree, tree_tmp_file.name, 'newick')
                ete_tree = Tree(open(tree_tmp_file.name).read(), format=1)
                children_tree = ete_tree.write(format=9)

        return children_tree

    # IBS calculation
    def _calculate_ibs(self, seqs, ids, sample1, sample2):

        index1 = ids.index(sample1)
        index2 = ids.index(sample2)
        ibs0 = 0
        ibs1 = 0
        ibs2 = 0
        for i, j in zip(seqs[index1], seqs[index2]):
            if i in ['N', '-'] or j in ['N', '-']:
                continue
            if i == j:
                ibs2 += 1
            else:
                ibs0 += 1
        ibs = (ibs2) / (ibs0+ibs2)

        return ibs

    # convert hierarchy to newick
    def _hier_to_newick(self, node, newick, leaf_names):

        if node.is_leaf():
            return '%s%s' % (leaf_names[node.id], newick)
        else:
            if len(newick) > 0:
                newick = ')%s' % (newick)
            else:
                newick = ');'
            newick = self._hier_to_newick(node.get_left(), newick, leaf_names)
            newick = self._hier_to_newick(node.get_right(), ',%s' % (newick), leaf_names)
            newick = '(%s' % (newick)

        return newick

    # function used for construction of write newick file
    def _write_nwk(self, ind_tree_info, samples):

        base = ''
        width, level, length =0, 0, 0.5
        trunk_list = []
        trunk_length = {}
        count_num = ind_tree_info[-1].count('-')
        for i in ind_tree_info[1:]:
            trunk = i.split('\t')[-1]
            width_new = i.count('-\t')
            if width_new > width:
                width = width_new
            level_new = i.count('-')-trunk.count('-')
            trunk_list.append(trunk)
            dif = level_new - level
            if dif == 0:
                length = length
                base = trunk + ',' + base
                if ind_tree_info.index(i)+1 == len(ind_tree_info):
                    base =  '('*(base.count(')')-base.count('(')) + base
            elif dif > 0:
                length = 1+1*(count_num-level_new)
                base = trunk + ')' + base
                if ind_tree_info.index(i)+1 == len(ind_tree_info):
                    base =  '('*(base.count(')')-base.count('(')) + base
            elif dif < 0:
                length = 1+1*(count_num-level_new)
                if ind_tree_info.index(i)+1 != len(ind_tree_info):
                    base =  trunk + ',' + '('*abs(dif) + base
                else:
                    base =  '('*(base.count(')')-base.count('(')-1) + trunk + ',(' + base
            trunk_length[trunk] = length
            level = level_new

        base = '(' + base + 'Y-Adam);'

        return base, trunk_length

    # plot tree figure
    def _visualize_tree(self, base, trunk_length, max_length, samples, layout, plot_info, color_dict=None):

        from ete3 import TreeStyle, NodeStyle, TextFace, RectFace
        # build tree from newick
        base_tree = Tree(base, format=8)

        num = 0
        for node in base_tree.traverse('postorder'):
            if node.name == '':
                node.name = 'BLANK_NODE%d' % num
            num += 1

        # add text in trunk and leaf, and keep all branch length same in topology
        for node in base_tree.traverse('postorder'):
            trunk = base_tree & node.name
            if node.name in samples:
                text = TextFace(node.name, ftype='Arial', fsize=18)
                text.margin_left = 4
                trunk.add_face(text, column=0, position='branch-right')
            elif node.name.startswith('BLANK_NODE'):
                text = ''
                trunk.add_face(TextFace(text, ftype='Arial', fsize=18), column=0, position='branch-top')
            else:
                space_num = max_length - len(node.name)
                text = node.name + ' '*space_num
                trunk.add_face(TextFace(text, ftype='Arial', fsize=18), column=0, position='branch-top')
            if node.is_leaf():
                node.dist = trunk_length[node.name]
            else:
                node.dist = 1
            gc.collect()

        # parameters for node style
        # the only difference between TreeNode.traverse() and TreeNode.iter_descendants() is that the first will include the root node in the iteration
        node_color = {}
        if isinstance(plot_info, pd.DataFrame): # match colors to population
            label = plot_info.columns[0]
            for node in base_tree.traverse('postorder'):
                ns = NodeStyle()
                ns['hz_line_width'] = 1
                ns['vt_line_width'] = 1
                if node.is_leaf() and node.name in samples:
                    # color
                    color = color_dict[plot_info.at[node.name, label]]
                    ns['fgcolor'] = color
                    node_color[node.name] = color
                    # shape
                    ns['shape'] = 'square'
                    ns['size'] = 50
                else:
                    children_nodes = [c_node for c_node in node.children]
                    node_colors = [node_color[i.name] for i in children_nodes]
                    if len(set(node_colors)) > 1:
                        if node.name in self.common_trunk or len(node.name) == 1:
                            color = 'black'
                        else:
                            for g_node in children_nodes:
                                node_colors += [node_color[i.name] for i in g_node]
                    node_color[node.name] = color
                    ns['fgcolor'] = 'black'
                    ns['size'] = 0
                ns['vt_line_color'] = color
                ns['hz_line_color'] = color
                node.set_style(ns)
        else: # all the leaf is same color
            for node in base_tree.traverse():
                ns = NodeStyle()
                ns['hz_line_width'] = 1
                ns['vt_line_width'] = 1
                ns['fgcolor'] = 'black'
                ns['size'] = 0
                node.set_style(ns)

        # parameters for tree style
        ts = TreeStyle()
        ts.min_leaf_separation = 30
        ts.show_leaf_name = False
        ts.show_branch_length = False
        ts.scale = max_length*15
        ts.branch_vertical_margin = 0
        ts.optimal_scale_level = 'full'
        ts.show_scale = False
        ts.margin_right = 20
        ts.margin_left = 20
        if layout == 'circular':
            ts.mode = 'c'

        # add legend and output tree
        base_tree.convert_to_ultrametric()
        if color_dict:
            if label == 'Color':
                # output figure with color information only
                base_tree.render(self.output_set[0]+'.pdf', tree_style=ts)
            else:
                # add legend
                for name, color in color_dict.items():
                    icon_legend = RectFace(width=200, height=200, fgcolor='black', bgcolor=color)
                    icon_legend.margin_left = 200
                    icon_legend.margin_right = 100
                    icon_legend.margin_top = 50
                    icon_legend.margin_bottom = 50
                    ts.legend.add_face(icon_legend, column=0)
                    text_legend = TextFace(name, fsize=200, ftype='Arial')
                    text_legend.margin_top = 50
                    text_legend.margin_bottom = 50
                    ts.legend.add_face(text_legend, column=1)
                ts.legend_position = 1
                # output figure with population information
                base_tree.render(self.output_set[0]+'.%s.pdf' % label, tree_style=ts)
        else:
            # output figure without population
            base_tree.render(self.output_set[0]+'.pdf', tree_style=ts)

    # match color to populations
    def _match_colors(self, population_info, label):

        # create a empty dict to match color and population
        color_dict = {}
        # if preset color, check color names
        if label == 'Color':
            colors = population_info[label]
            for i in set(colors):
                if matplotlib.colors.is_color_like(i):
                    color_dict[i] = i
                else:
                    self.logger.error('[Y-LineageTracker] [Error] %s is not a RGB color' % i)
                    sys.exit()
        else:
            # match color to populations
            populations = sorted(list(set(population_info[label])))
            from FilesIO import set_color_num
            color_num = len(populations)
            colors, cmap = set_color_num(color_num, get_cmap=True)

            # creat a dict usef for match color to population
            for i, j in zip(populations, cmap):
                color_dict[i] = matplotlib.colors.to_hex(j)

        plot_info = population_info[[label]]

        return plot_info, color_dict

    # main function for tree plot
    def output_tree_figure(self, population_info, base, trunk_length, max_length, samples, layout):

        if isinstance(population_info, pd.DataFrame):
            if 'Color' in population_info.columns:
                plot_info, color_dict = self._match_colors(population_info, 'Color')
                self._visualize_tree(base, trunk_length, max_length, samples, layout, plot_info, color_dict)
            else:
                if 'Population' in population_info.columns:
                    plot_info, color_dict = self._match_colors(population_info, 'Population')
                    self._visualize_tree(base, trunk_length, max_length, samples, layout, plot_info, color_dict)
                if 'Group' in population_info.columns:
                    plot_info, color_dict = self._match_colors(population_info, 'Group')
                    self._visualize_tree(base, trunk_length, max_length, samples, layout, plot_info, color_dict)
        else:
            plot_info = None
            self._visualize_tree(base, trunk_length, max_length, samples, plot_info, layout)


def main():

    start = time.perf_counter()
    arguments = phylo_parser()

    # set of output files
    output_set = check_phylo_output(arguments.tree, arguments.output)

    from FilesIO import check_hap_input, check_population, check_overlap, time_count
    # check haplogroup data
    hap_data = check_hap_input(arguments.hg, 'haplogroup')

    # check population data
    if arguments.population:
        population_data = check_population(arguments.population)
        hap_data, population_data = check_overlap(hap_data, population_data)

    # set log file
    set_log(arguments, output_set[-1])

    # start phylogeny analysis
    phylo = PhyloHaplogroup(hap_data, output_set)
    # read sequence alignment file
    if arguments.seq:
        ids, seqs = phylo.read_seq_file(arguments.seq, arguments.format)
        hap_df, base, trunk_length, max_length, samples = phylo.get_base_tree(ids) # build base tree
    else:
        hap_df, base, trunk_length, max_length, samples = phylo.get_base_tree()
    # construct sub-tree internal nodes
    if arguments.seq:
        base = phylo.phylo_internal(ids, seqs, base, samples, arguments.align)
    # combine with population data
    if arguments.population:
        population_info = pd.merge(hap_df, population_data, on='SampleID')
        population_info = population_info.set_index('SampleID')
    else:
        population_info = None
    # output tree in specific format
    phylo.output_tree(base, arguments.tree)
    # output tree figure
    phylo.output_tree_figure(population_info, base, trunk_length, max_length, samples, arguments.layout)

    time_count(start)


if __name__ == '__main__':
    main()



import os
import sys
import time
import logging
import argparse
import matplotlib; matplotlib.use('Agg')
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from itertools import filterfalse
from collections import Counter
from matplotlib import font_manager
from GetConfig import getConfig

try:
    from adjustText import adjust_text
except ImportError:
    print('Module [adjustText] is required, please install it by command line: pip install adjustText')
    sys.exit()


'''
Description:
This module is used for clustering analysis.
Clustering method can be applied by --method option.
The program will take haplogroup file and population file as input files.
Then cluster populations by PCA or MDS method and visualize results by the plot.
'''


config = getConfig() # get config


def cluster_parser():

    parser = argparse.ArgumentParser('cluster', description='(c) Y-LineageTracker: NRY haplogroup clustering analysis')
    # function used for clustering analysis
    parser.add_argument('cluster',
                        help='Pefrorm clustering analysis for NRY haplogroups.')
    # required, haplogroup file
    parser.add_argument('--hg',
                        required=True,
                        type=str,
                        action='store',
                        help='hg: A file containing sample ID and haplogroup of each individual.')
    # required, population file
    parser.add_argument('-p', '--population',
                        required=True,
                        type=str,
                        action='store',
                        help='population: A file containing sample ID and population information of each individual.')
    # optional, clustering algorithm
    parser.add_argument('--method',
                        required=False,
                        type=str,
                        action='store',
                        default='pca',
                        choices=['pca', 'mds'],
                        help='method: The algorithm used for clustering analysis (PCA or MDS).')
    # optional, simpification mode for haplogroups in same main trunk
    parser.add_argument('--level',
                        required=False,
                        type=str,
                        const='auto',
                        action='store',
                        nargs='?',
                        help='Level: the haplogroup resolution level of NRY haplogroups for clustering,\
                              which means the degree of similarity among haplogroups in the same main trunk.\
                              The parameter [auto] will set an optimal level to simplify NRY haplogroups resolution.')
    # optional, output haplogroup frequence of populations
    parser.add_argument('--freq',
                        required=False,
                        action='store_true',
                        help='frequence: Output the haplogroup frequency of each population.')
    # optional, the prefix of output
    parser.add_argument('-o', '--output',
                        required=False,
                        type=str,
                        action='store',
                        help='output: The prefix of output files.')


    args = parser.parse_args()

    return args


# print program information and write to log file
def set_log(log_file, level, args_log):

    logger = logging.getLogger()
    logger.setLevel(level=logging.INFO)

    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] - [%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    logger.addHandler(handler)
    logger.addHandler(console)

    log_info = ['[Y-LineageTracker] [Cluster]',
                '[Y-LineageTracker] Run Date: ' + time.asctime(time.localtime(time.time())),
                '[Y-LineageTracker] Input File: %s' % args_log.hg,
                '[Y-LineageTracker] Clustering Method: %s' % args_log.method,
                '[Y-LineageTracker] Haplogroup Level: %s' % level]

    if args_log.freq:
        log_info.append('[Y-LineageTracker] Output haplogroup frequence')

    print('\n')
    for i in log_info:
        logger.info(i)


# names of output files
def check_cluster_output(output, method, freq):

    from FilesIO import get_out_path
    path = get_out_path(output, 'Cluster')
    level_file = path + '.level.hg'
    if method == 'pca':
        fig_file = path + '.PCA.fig.pdf'
        vect_file = path + '.PCA.eigenvectors.txt'
        val_file = path + '.PCA.eigenval.txt'
        output_set = [fig_file, vect_file , val_file, level_file]
        log_file = path + '.ClusterLog.PCA.log'
    elif method == 'mds':
        fig_file = path + '.MDS.fig.pdf'
        fst_file = path + '.MDS.fst.txt'
        embedding_file = path + '.MDS.embedding.txt'
        output_set = [fig_file, fst_file, embedding_file, level_file]
        log_file = path + '.ClusterLog.MDS.log'
    if freq:
        freq_file = path + '.freq.txt'
        output_set.append(freq_file)

    return output_set, log_file


# function of haplogroup simplification
def hap_simplify(haplogroup, level):

    if haplogroup.startswith('A'):
        if haplogroup.startswith('A1'):
            return haplogroup[:level+1]
        elif haplogroup.startswith('A0'):
            zero_num = haplogroup.count('0')
            if 'T' in haplogroup:
                return haplogroup[:level+zero_num+1]
            else:
                return haplogroup[:level+zero_num]
    else:
        if (haplogroup.isalpha() and haplogroup.isupper()):
            return haplogroup
        elif haplogroup[:-1].isupper and haplogroup[:-1].isalpha() and len(haplogroup) > 2:
            return haplogroup
        else:
            return haplogroup[:level]

# check level parameter
def check_level(level):

    if level:
        if level in ['auto', 'original', 'min']:
            return level
        else:
            try:
                level = int(level)
                if level >= 1:
                    return level
                else:
                    raise argparse.ArgumentTypeError('parameter [level] should be an integer or auto')
            except ValueError:
                raise argparse.ArgumentTypeError('parameter [level] should be an integer or auto')
    else:
        level = 'original'
        return level


class ClusterAnalysis(object):
    '''
    This class is used to perform main clustering analysis
    Four main steps are included in this class:
    1. read haplogroup file and simplify NRY haplogroups
    2. read population file and merge to haplogroup data
    3. perform clustering analysis
    4. output clustering result with figure
    '''

    def __init__(self):

        self.logger = logging.getLogger()

    # calculate Fst
    # Weir and Hill. 2002. Estimating F-statistics. Ann.Rev.Gen.
    def _calculate_pairwise_fst(self, pop1_count, pop2_count):

        sample_size1 = (np.sum(pop1_count)/pop1_count.size)*2
        sample_size2 = (np.sum(pop2_count)/pop2_count.size)*2

        pop1_frequency = np.array(pop1_count) / sample_size1
        pop2_frequency = np.array(pop2_count) / sample_size2

        mean_freq = (np.array(pop1_count)+np.array(pop2_count)) / (sample_size1+sample_size2)

        MSP1 = sample_size1*((pop1_frequency-mean_freq)**2)
        MSP2 = sample_size2*((pop2_frequency-mean_freq)**2)
        MSP = MSP1 + MSP2

        MSG1 = float(1/((sample_size1-1)+(sample_size2-1)))
        MSG2 = (np.array(pop1_count) * (1-pop1_frequency)) + (np.array(pop2_count) * (1-pop2_frequency))
        MSG = MSG1 * MSG2

        nc = sample_size1+sample_size2 - (sample_size1**2+sample_size2**2)/(sample_size1+sample_size2)

        theta1 = np.sum(MSP-MSG)
        theta2 = np.sum(MSP+(nc-1)*MSG)
        if theta2 == 0:
            theta = 0
        else:
            theta = theta1 / theta2

        return theta

    # calculate pairwise Fst
    def _calculate_MDS_fst(self, population_count, fst_output):

        pops = population_count.index.tolist()
        fst_df = pd.DataFrame(index=pops, columns=pops)
        pop_num = 0
        for pop1 in pops:
            for pop2 in pops[pop_num:]:
                if pop1 == pop2:
                    fst_df.at[pop2, pop1] = 0
                else:
                    pop1_count = np.empty(shape=(2, population_count.columns.size))
                    pop1_count[0] = np.array(population_count.loc[pop1].tolist()).reshape(1, -1)
                    pop1_count[1] = (np.sum(pop1_count[0])-np.array(population_count.loc[pop1].tolist())).reshape(1, -1)
                    pop2_count = np.empty(shape=(2, population_count.columns.size))
                    pop2_count[0] = np.array(population_count.loc[pop2].tolist()).reshape(1, -1)
                    pop2_count[1] = (np.sum(pop2_count[0])-np.array(population_count.loc[pop2].tolist())).reshape(1, -1)
                    fst = self._calculate_pairwise_fst(pop1_count, pop2_count)
                    fst_df.at[pop2, pop1] = fst
                    fst_df.at[pop1, pop2] = fst
            pop_num += 1

        fst_df.to_csv(fst_output, sep='\t', index=True)
        fst_array = np.array(fst_df)

        return fst_array

    # simplify haplogroups according to level
    def classify_hap_data(self, level, hap_data, level_file):

        if level != 'original':
            # simplify to most common haplogroup
            if level == 'auto':
                initial = set(hap_data['Haplogroup'].map(lambda x: x[0]))
                common_list = []
                for i in initial:
                    initial_group = sorted(list(filter(lambda x: x[0] == i, hap_data['Haplogroup'])), key=lambda x: len(x))
                    while True:
                        most_common = Counter(initial_group).most_common(1)[0][0]
                        initial_group = [most_common if x.startswith(most_common) and len(x) > len(most_common) else x for x in initial_group]
                        hap_num = len(most_common)-1
                        for j in range(len(most_common)-1):
                            hap = most_common[0:hap_num]
                            if hap in initial_group:
                                initial_group = [hap if x == most_common else x for x in initial_group]
                                most_common = hap
                            hap_num -= 1
                        common_list.append(most_common)
                        initial_group = list(filterfalse(lambda x: x.startswith(most_common), initial_group))
                        if len(initial_group) == 0:
                            break
                common_list = sorted(common_list, key=lambda x: len(x), reverse=True)
                for i in range(len(hap_data['Haplogroup'])):
                    for j in common_list:
                        if hap_data.at[i, 'Haplogroup'].startswith(j):
                            hap_data.at[i, 'Haplogroup'] = j
            # simplify to specific level
            else:
                if level == 'min':
                    main_trunk = config.get('HaplogroupTree', 'CommonTrunk').split(',') + config.get('HaplogroupTree', 'UpperTrunk').split(',')
                    level = min(map(lambda x: len(x), fiter(lambda x: x not in main_trunk, hap_data['Haplogroup'])))
                hap_data['Haplogroup'] = hap_data['Haplogroup'].apply(hap_simplify, level=level)
            hap_data.to_csv(level_file, sep='\t', index=False)


        return hap_data

    # read population data and plot clustering figure
    def output_cluster(self, method, freq, output_set, hap_data, population_data):

        #############################READ POPULATION DATA#############################

        # merge population data for frequence construction
        hap_data = pd.merge(hap_data, population_data, on='SampleID')

        # count number of populations and sort them in order
        num_list = []
        num = 0
        for i in sorted(list(set(population_data['Population'])), key=population_data.drop('SampleID', axis=1).drop_duplicates()['Population'].tolist().index):
            num_list.append((num, num+population_data['Population'].tolist().count(i)))
            num = num + population_data['Population'].tolist().count(i)
        population_data = population_data.drop('SampleID', axis=1).drop_duplicates()

        # sort according to Group for subsequent plot
        if 'Group' in population_data.columns:
            population_data = population_data.sort_values(by='Group')

        # haplogorup frequence of populations
        population_freq = pd.DataFrame(index=sorted(list(set(hap_data['Population'])),
                                       key=population_data['Population'].tolist().index),
                                       columns=sorted(list(set(hap_data['Haplogroup']))),
                                       dtype=np.float)
        hap_series = pd.Series(hap_data['Haplogroup'].tolist(), index=[hap_data['Population']])
        if method == 'mds':
            population_count = pd.DataFrame(index=sorted(list(set(hap_data['Population'])),
                                            key=population_data['Population'].tolist().index),
                                            columns=sorted(list(set(hap_data['Haplogroup']))),
                                            dtype=np.int)
        # count haplogroup frequence
        for i in population_freq.index:
            hap_list = hap_series[i].value_counts().index.tolist()
            freq_list = hap_series[i].value_counts().tolist()
            sum_num = sum(freq_list)
            for j in range(len(hap_list)):
                population_freq.at[i, hap_list[j]] = freq_list[j] / sum_num
                if method == 'mds':
                    population_count.at[i, hap_list[j]] = freq_list[j]
        population_freq = population_freq .fillna(0)
        if method == 'mds':
            population_count = population_count.fillna(0)

        #############################PERFORM CLUSTERING ALGORITHM#############################

        matrix = population_freq.to_numpy() # convert haplogroup frequency dataframe to martix
        if method == 'pca': # PCA analysis
            from sklearn.decomposition import PCA
            pca = PCA(copy=True, whiten=False)
            coordinates = pca.fit_transform(matrix)
            # output PCA results
            output_vect_file = output_set[1]
            output_vect_data = pd.DataFrame(coordinates, index=sorted(list(set(hap_data['Population']))))
            output_vect_data.to_csv(output_vect_file , sep='\t', index=True)
            output_val_file = output_set[2]
            output_val_data = pd.DataFrame(pca.explained_variance_)
            output_val_data.to_csv(output_val_file, index=False, header=None)
        elif method == 'mds': # MDS analysis
            from sklearn.manifold import MDS
            mds_matrix = self._calculate_MDS_fst(population_count, output_set[1])
            mds = MDS(dissimilarity='precomputed')
            coordinates = mds.fit_transform(mds_matrix)
            # output MDS result
            output_embedding_file = output_set[2]
            output_embedding_data = pd.DataFrame(coordinates, index=sorted(list(set(hap_data['Population']))))
            output_embedding_data.to_csv(output_embedding_file , sep='\t', index=True)
        # output frequence of haplogroup in populations
        if freq:
            output_freq_file = output_set[-1]
            population_freq.sort_index().to_csv(output_freq_file, sep='\t', index=True)

        #############################START PLOT#############################

        # defualt color and shape in cluster if no Group columns in population file
        color_list = ['slategrey']*population_data.index.size
        shape_list = ['o']*population_data.index.size

        # set color according to Group labels
        if 'Group' in population_data.columns:
            label_list = population_data['Group'].tolist()
            # if with color columns
            if 'Color' in population_data.columns:
                color_list = population_data['Color'].tolist()
            # get cmap colors in matplotlib
            else:
                from FilesIO import set_color_num
                color_num = len(set(label_list))
                colors = set_color_num(color_num)
                label_count = 0
                color_list = []
                colored_label = []
                label_dict = {}
                for i in label_list:
                    if i not in colored_label:
                        label_dict[i] = colors[label_count]
                        colored_label.append(i)
                        color_list.append(label_dict[i])
                        label_count += 1
                    else:
                        color_list.append(label_dict[i])
        # No Group columns
        elif 'Color' in population_data.columns and 'Group' not in population_data.columns:
            color_list = population_data['Color'].tolist()
        if 'Shape' in population_data.columns:
            shape_list = population_data['Shape'].tolist()

        self.logger.info('[Y-LineageTracker] Start clustering analysis')

        # parameters for plot
        import warnings
        warnings.filterwarnings('ignore')
        font_files = font_manager.findSystemFonts(fontpaths=[os.path.split(os.path.realpath(__file__))[0] + '/sans-serif'])
        font_list = font_manager.createFontList(font_files)
        font_manager.fontManager.ttflist.extend(font_list)
        plt.rcParams['font.family']= 'sans-serif'
        plt.rcParams['font.sans-serif'] = ['Arial'] # font
        plt.rcParams['axes.unicode_minus'] = False
        figure, ax = plt.subplots(figsize = (12, 12)) # figure size

        # x and y lines of zero
        ax.axvline(x=0, color='gray', linestyle='--', linewidth=1.5, alpha=0.7)
        ax.axhline(y=0, color='gray', linestyle='--', linewidth=1.5, alpha=0.7)

        # x and y labels
        if method == 'pca':
            plt.xlabel('PC1  %.2f%%'% (pca.explained_variance_ratio_[0]*100), fontsize=25)
            plt.ylabel('PC2  %.2f%%'% (pca.explained_variance_ratio_[1]*100), fontsize=25)
        elif method == 'mds':
            plt.xlabel('Dimension1', fontsize=25)
            plt.ylabel('Dimension2', fontsize=25)

        plt.xticks(size=20)
        plt.yticks(size=20)

        # traverse populations to plot scatterplot
        if 'label_list' in vars():
            plot_list = []
            for i in range(population_data.index.size):
                plot = list(zip(color_list, shape_list, label_list))[i]
                if plot in plot_list:
                    plt.plot(coordinates[i, 0], coordinates[i, 1], shape_list[i],
                             color=color_list[i],
                             mec='black',
                             markersize=15,
                             alpha=0.9)
                else:
                    plt.plot(coordinates[i, 0], coordinates[i, 1], shape_list[i],
                             color=color_list[i],
                             mec='black',
                             label=label_list[i],
                             markersize=15,
                             alpha=0.9)
                plot_list.append(plot)
            plt.legend(frameon=True, fontsize=15, markerscale=1, shadow=True, edgecolor='dimgrey', loc='best')
        else:
            for i in range(population_data.index.size):
                plt.plot(coordinates[num_list[i][0]:num_list[i][1], 0], coordinates[num_list[i][0]:num_list[i][1], 1], shape_list[i],
                         color=color_list[i],
                         mec='black',
                         markersize=15,
                         alpha=0.7)

        # plot population with labels
        if population_freq.index.size < 40:
            texts = [plt.text(coordinates[i, 0], coordinates[i, 1], s=population_freq.index.tolist()[i], size=15) for i in range(len(matrix))]
            import warnings
            warnings.filterwarnings('ignore')
            adjust_text(texts)
        else:
            self.logger.warning('[Y-LineageTracker] Too much population in the population file')

        # output figure
        output_fig_file = output_set[0]
        plt.savefig(output_fig_file, dpi=100, bbox_inches='tight')
        plt.close()

        self.logger.info('[Y-LineageTracker] Clustering analysis finished')


def main():

    start = time.perf_counter()
    arguments = cluster_parser()

    # check haplogroup level
    global level
    level = check_level(arguments.level)

    # set of output files
    output_set, log_file = check_cluster_output(arguments.output, arguments.method, arguments.freq)

    # set log file
    set_log(log_file, level, arguments)

    from FilesIO import check_hap_input, check_population, check_overlap, time_count
    # check haplogroup data
    hap_data = check_hap_input(arguments.hg, 'haplogroup')

    # check population data
    population_data = check_population(arguments.population)

    # check overlap of haplogroup data and population data
    hap_data, population_data = check_overlap(hap_data, population_data)

    # start cluserting analysis
    cluster = ClusterAnalysis()
    hap_data = cluster.classify_hap_data(level, hap_data, output_set[3])
    cluster.output_cluster(arguments.method, arguments.freq, output_set, hap_data, population_data)

    time_count(start)


if __name__ == '__main__':
    main()



import re
import sys
import logging
import numpy as np
import pandas as pd

from functools import reduce
from itertools import filterfalse
from collections import Counter, OrderedDict
from GetConfig import getConfig


config = getConfig()


class TrackLineage(object):
    '''
    This class is used to:
    1. classify NRY haplogroup
    2. Track lineage
    3. Output result of classification
    '''

    def __init__(self, matched_info, ref_ind_info, approximate_haps, build, format, chip):

        # record running log
        self.logger = logging.getLogger()

        # get reference haplogroups of which all indivisuals are same
        if build == '37' or 'Ref':
            idx = -7
        elif build == '38':
            idx = -8

        # get parameters from config file
        self.build = build
        self.marker_rate = float(config.get('HaplogroupTree', 'MarkerRate'))
        self.length_cutoff = config.getint('HaplogroupTree', 'LengthCutoff')
        self.matched_marker = config.get('HaplogroupMarker', 'MatchMarker')
        self.match_miss_marker = config.get('HaplogroupMarker', 'MatchMissMarker')
        self.track_symbol = config.get('HaplogroupMarker', 'TrackSymbol') # track symbol is '->'
        self.track_rate = float(config.get('HaplogroupTree', 'TrackRate'))
        self.tree_trunk = config.get('HaplogroupTree', 'CommonTrunk').split(',')
        self.approximate_haps = approximate_haps
        self.haplogroup_interval = config.getint('HaplogroupTree', 'HaplogroupInterval')

        if format == 'vcf' and not chip and build != 'Ref':
            common_sites = [i for i in ref_ind_info['Build'+str(build)].tolist() if i not in matched_info[:, idx].tolist()]
            common_df = ref_ind_info[ref_ind_info['Build'+str(build)].map(lambda x: x in common_sites)]
            self.common_haps = list(set(common_df['Haplogroup'])&set(self.tree_trunk))
            self.common_key_haps = list(set(common_df[common_df['KeyInfo']=='Key']['Haplogroup'])&set(self.tree_trunk))
        else:
            self.common_key_haps = []
            self.common_haps = []

    # extract matched haplogroups and their frequence, individuals with no matched item will be removed and recoreded
    def _extract_haplogroup(self, col, key_matched_info, ref_haps):

        # extract all matched key haplogroups
        key_haps = list(map(lambda x: x[1:], list(filter(lambda x: x[0] == self.matched_marker, key_matched_info[:, col])))) + self.common_key_haps
        count_res = OrderedDict(sorted(Counter(all_matched_haps+self.common_haps).items()))
        matched_haps = list(count_res.keys())
        matched_haps_freq = list(count_res.values())
        total_haps_freq = list(dict(filter(lambda x: x[0] in set(matched_haps), ref_haps)).values())

        unzipped_haps = [matched_haps, matched_haps_freq, total_haps_freq]

        return key_haps, unzipped_haps

    # find the most downstream haplogroup according to key haplogroups
    def _search_final_hap(self, key_haps):

        # tracker rate of haplogroup A is special since the complex names
        haplogroup_A_rate = float(config.get('HaplogroupTree', 'HaplogroupARate'))

        # set initial haplogroup is NONE
        key_hap = self.match_miss_marker
        final_hap = ''

        # traverse to find a most reliable key haplogroup
        length = 0
        for hap in filterfalse(lambda x: x.startswith('K') or x in self.tree_trunk, key_haps):
            length2 = len(hap)
            if length2 == 1:
                continue
            if length2 >= length:
                exist_num = 0
                hap_num = length2 - 1
                for i in hap:
                    hap_str = hap[0: hap_num]
                    if hap_str in key_haps:
                        exist_num += 1
                    hap_num -= 1
                if exist_num == length2-1:
                    length = length2
                    final_hap = hap
                    continue
                if length2 >= self.length_cutoff and (exist_num+1)/length2 > self.track_rate:
                    length = length2
                    final_hap = hap
                    continue
                if hap.startswith('A'):
                    if (exist_num+1)/length2 > (self.track_rate - haplogroup_A_rate):
                        length = length2
                        final_hap = hap
                    elif hap == 'A0' and 'A0T' in key_haps:
                        final_hap = hap
                        length = length2
                    elif hap.startswith('A1b') and 'A1b' in key_haps and 'A1' in key_haps:
                        final_hap = hap
                        length = length2

        return key_hap, final_hap

    # search whether there are more downstream haplogroup
    def _search_downstream(self, matched_haps, zipped_haps, key_haps, key_hap, final_hap):

        key_hap = final_hap
        length = len(final_hap)
        rate = 0
        downstream = sorted(filter(lambda x: len(x[0]) >= len(final_hap)+1 and
                                                 x[0].startswith(final_hap) and
                                                 x[0] not in self.tree_trunk, zipped_haps),
                                                 key=lambda x: len(x[0]))

        traversed = []
        for hap, matched, total in downstream:
            length = len(hap)
            final_hap = hap
            rate = matched/total
            if final_hap in key_haps and len(hap)-length < self.haplogroup_interval:
                key_hap = final_hap
            else:
                if matched/total > rate and len(hap)-length < self.haplogroup_interval:
                    length = len(hap)
                    final_hap = hap
                    rate = matched/total
                    if final_hap in key_haps:
                        key_hap = final_hap
                elif len(list(filter(lambda x: hap.startswith(x), traversed))) > len(list(filter(lambda x: final_hap.startswith(x), traversed)))-1: # exclude final hap itself
                    length = len(hap)
                    final_hap = hap
                    rate = matched/total
                    if final_hap in key_haps:
                        key_hap = final_hap
            traversed.append(hap)

        return key_hap, final_hap

    # if the final haplogroup is in K branch
    def _replace_single_K_trunk_haplogroup(self, key_haps, key_hap):

        if key_hap == 'K2b1' or key_hap == 'K2b':
            if 'S' in key_haps:
                key_hap = 'S'
            elif 'M' in key_haps:
                key_hap = 'M'
            elif 'P1' in key_haps:
                key_hap = 'P1'
            elif 'P' in key_haps:
                key_hap = 'P'
        elif key_hap == 'K2':
            if 'NO1' in key_haps:
                key_hap = 'NO1'
            elif 'NO' in key_haps:
                key_hap = 'NO'

        final_hap = key_hap

        return key_hap, final_hap

    # if the resolution of data is low, detect potential haplogroups
    def _detect_low_resolution_haplogroup(self, zipped_haps, matched_haps, key_haps, final_hap, key_hap='.'):

        length = 0
        rate = 0
        flag = False
        downstream = sorted(zipped_haps, key=lambda zipped_haps: len(zipped_haps[0]))
        downstream = filter(lambda x: x[0] not in self.tree_trunk, downstream)
        for hap, matched, total in downstream:
            if hap[0] not in matched_haps:
                continue
            if matched/total <= (self.marker_rate-0.45):
                continue
            if hap.startswith('A'):
                matched_haps.append('A')
            length2 = len(hap)
            if (length2 > length) or (length2 == length and matched/total > rate):
                exist_num = 0
                hap_num = length2 - 1
                for i in hap:
                    hap_str = hap[0: hap_num]
                    if hap_str in matched_haps:
                        exist_num += 1
                    hap_num -= 1
                if exist_num == length2-1:
                    flag = True
                else:
                    if length2 >= 4 and (exist_num+1)/length2 > (self.track_rate-0.2):
                        flag  = True
                if flag:
                    length = length2
                    final_hap = hap
                    rate = matched/total
                    flag = False
                    for j in reversed(sorted(key_haps)):
                        if final_hap.startswith(j):
                            key_hap = j
                            break

        if final_hap:
            key_hap, final_hap = self._search_downstream(matched_haps, zipped_haps, key_haps, key_hap, final_hap)

        return key_hap, final_hap

    # if the resolution of data is lower than above, detect potential haplogroups
    def _detect_lower_resolution_haplogroup(self, zipped_haps, key_haps, key_hap):

        if len(list(filterfalse(lambda x: x in self.tree_trunk, key_haps))) > 1:
            length = 0
            last_haps = list(filterfalse(lambda x: x in self.tree_trunk, key_haps))
            for hap in last_haps:
                length2 = len(hap)
                if length2 >= length:
                    final_hap = hap
                    length = length2
                    key_hap = self.matched_marker + final_hap
        else:
            hap_rate = 0
            last_haps = sorted(zipped_haps, key=lambda zipped_haps: zipped_haps[0])
            last_haps = sorted(last_haps, key=lambda last_haps: len(zipped_haps[0]))
            if set(list(zip(*last_haps))[0]).issubset(set(self.tree_trunk)):
                final_hap = last_haps[-1][0]
            else:
                for hap, matched, total in last_haps:
                    if len(hap) > 1 and hap in self.tree_trunk:
                        continue
                    if matched/total >= hap_rate:
                        final_hap = hap
                        hap_rate = matched/total

        return key_hap, final_hap

    # get a lineage track according to ISOGG tree information
    def _track_lineage(self, final_hap):

        from FilesIO import CommonData
        common_data = CommonData()
        tree_info = common_data.read_tree_info()

        if (final_hap.isalpha() and final_hap.isupper()) or final_hap == 'NO1':
            track = final_hap
        else:
            track = ''
            alpha = not final_hap[-1].isalpha()
            for num_str in range(len(final_hap)):
                tmp_up = final_hap[0: len(final_hap)-num_str]
                is_alpha = tmp_up[-1].isalpha()
                if is_alpha != alpha:
                    up = tmp_up
                else:
                    continue
                if up in self.approximate_haps:
                    up = up + '~'
                track = track + self.track_symbol + up
                alpha = is_alpha
            track = track[2:]
            # A is special haplogroup in name
            if track[-1] == config.get('HaplogroupA', 'SpecificHaplogroup'):
                track_temp = track[:-3]
                track = self.track_symbol + track_temp + config.get('HaplogroupA', 'SpecificTrackOne')
                pattern_track = re.compile(r'A00[abc]')
                if re.search(pattern_track, track):
                    track_temp = pattern_track.search(track).group()
                    track = self.track_symbol + track_temp + config.get('HaplogroupA', 'SpecificTrackTwo')
                if config.get('HaplogroupA', 'SpecificSubhaplogroup') in track:
                    track = config.get('HaplogroupA', 'SpecificTrackThree')

        pattern = re.compile(r'[A-Za-z0-9]+')
        length = 0
        num = 0
        header = track.split('>')[-1]
        for i in tree_info:
            if len(i) < len(header)+1:
                continue
            if i[-(len(header)+1)] == '\t' and header == i[-len(header):]:
                    length = i.count('-')
                    for j in list(reversed(tree_info)):
                        if j.count('-') + 1 == length and tree_info.index(j) < tree_info.index(i):
                            searched = pattern.search(j).group()
                            track = track + self.track_symbol + searched
                            length = length - 1
                        else:
                            continue
                    track = self.track_symbol + track + self.track_symbol + 'Y-Adam'

        return track

    # count how many mutations matched
    def _count_matched_number(self, track, unzipped_haps):

        matched_haps, matched_haps_freq, total_haps_freq = unzipped_haps
        haps_in_track = track.split(self.track_symbol)[:-1]
        for hap in haps_in_track:
            if hap in matched_haps:
                hap_index = matched_haps.index(hap)
            elif hap.strip('~') in matched_haps:
                hap_index = matched_haps.index(hap.strip('~'))
            else:
                continue
            matched = matched_haps_freq[hap_index]
            total = total_haps_freq[hap_index]
            hap_without_count = self.track_symbol + hap + self.track_symbol
            hap_with_count = self.track_symbol + hap + '(' + str(matched) + '/' + str(total) + ')' + self.track_symbol
            track = track.replace(hap_without_count, hap_with_count)
        track = track[2:]

        return track

    # main function of classification: classify haplogroup of each individual
    def classify_haplogroup(self, inds, matched_info, ref_info, output_set, mutation, header_cut):

        # output data one: haplogroup classification result
        hap_data = pd.DataFrame(index=inds,
                                columns=['Haplogroup'])
        output_hap_file= output_set[0]

        # output data two: lineage of haplogroups
        lineage_data = pd.DataFrame(index=inds,
                                    columns=['Haplogroup',
                                            'KeyHaplogroup',
                                            'Mutations',
                                            'LineageTrack'])
        output_lineage_file = output_set[1]
        # output data three: all matched mutation
        if mutation:
            mutation_data = pd.DataFrame(index=inds,
                                         columns=['AllMutations'])
            output_mutation_file = output_set[2]

        # initial variables
        no_hap_inds = []

        # get freq of all total matched haplogroups
        ref_haps = sorted(Counter(ref_info['Haplogroup']).items())

        base_cols = list(range(0, 9)) + list(range(-11, 0))
        base_info = matched_info[:, base_cols]
        key_matched_info = matched_info[matched_info[:,-4]=='Key']

        # traverse all indivisual to classify haplogroups
        final_list = []
        key_list = []
        mutation_list = []
        track_list = []
        all_mutation_list = []
        ind_num = 0
        for n, ind in enumerate(inds):

            col = n + header_cut
            col_info = matched_info[:, col]

            # extract all matched haplogroups
            global all_matched_haps
            # [1:] means remove match marker
            all_matched_haps = sorted(list(map(lambda x: x[1:], list(filter(lambda x: x[0]==self.matched_marker, col_info)))))

            # whether there are individual who don't have matched haplogroups
            if len(all_matched_haps) == 0:
                no_hap_inds.append(ind)
                final_list.append('.')
                key_list.append('.')
                mutation_list.append('.')
                track_list.append('.')
                self.logger.warning('[Y-LineageTracker] [Warning] There are no mutation matched for sample %s' % (ind))
                continue

            # extract matched haplogroups and their frequence
            key_haps, unzipped_haps = self._extract_haplogroup(col, key_matched_info, ref_haps)
            matched_haps, matched_haps_freq, total_haps_freq = unzipped_haps

            zipped_haps = list(zip(*unzipped_haps))

            if len(key_haps) > 0: # if there are key haplogroups
                key_hap, final_hap = self._search_final_hap(key_haps) # whether there is a key haplogroup, data with low resolution will get an empty string
                if final_hap: # haplogroup with single character will keep an empty string
                    key_hap, final_hap = self._search_downstream(matched_haps, zipped_haps, key_haps, key_hap, final_hap) # whether the final haplogroup have a more donwstream one
                else: # there are key haplogroups but not classified
                    key_hap, final_hap = self._detect_low_resolution_haplogroup(zipped_haps, matched_haps, key_haps, final_hap, key_hap)
            else: # no key haplogroup matched
                if self.build != 'Ref':
                    self.logger.warning('[Y-LineageTracker] [Warning] There are no key haplogroups matched for sample %s' % (ind))
                final_hap = ''
                key_hap, final_hap = self._detect_low_resolution_haplogroup(zipped_haps, matched_haps, key_haps, final_hap)

            if key_hap.startswith('K'): # other haplogroups not started with K have been tested
                key_hap, final_hap = self._replace_single_K_trunk_haplogroup(key_haps, key_hap)

            if final_hap == '': # if the haplogroup is None
                key_hap, final_hap = self._detect_lower_resolution_haplogroup(zipped_haps, key_haps, key_hap)

            # match all matched haplogroups to their mutations
            matched_haps_array = base_info[list(map(lambda x: x[0] == self.matched_marker, col_info))]
            mutations = ', '.join(matched_haps_array[:, -11])

            # match final haplogroup to its mutations
            final_hap_array = matched_haps_array[matched_haps_array[:, -10] == final_hap]
            final_mutations = ', '.join(final_hap_array[:, -11])
            if 'Key' not in final_hap_array[:, -4]:
                final_mutations = '*' + final_mutations

            # get lineage track of haplogroup
            track = self._track_lineage(final_hap)

            # count matched number of each haplogroup
            haps_in_track = track.split(self.track_symbol)[:-1]
            track = self._count_matched_number(track, unzipped_haps)

            # output row for each individual
            mutation_list.append(final_mutations)
            track_list.append(track)
            if mutation:
                all_mutation_list.append(mutations)
            if final_hap in self.approximate_haps:
                final_hap = final_hap + '~'

            final_list.append(final_hap)
            key_list.append(key_hap)

            percent = float(ind_num+1)*100 / float(len(inds))
            sys.stdout.write('[Y-LineageTracker] Classifying haplogroups…… %.2f' % percent)
            sys.stdout.write('%\r')
            sys.stdout.flush()
            ind_num += 1

        hap_data['Haplogroup'] = final_list
        # output file one: haplogroup classification result
        hap_data.index.name = 'SampleID' # set individuals as index
        hap_data.to_csv(output_hap_file, sep='\t', index=True, encoding='utf-8')

        print()
        self.logger.info('[Y-LineageTracker] Classification and Tracking haplogroup lineages finished')

        # output file two: lineage of each haplogroup
        lineage_data['Haplogroup'] = final_list
        lineage_data['KeyHaplogroup'] = key_list
        lineage_data['Mutations'] = mutation_list
        lineage_data['LineageTrack'] = track_list
        lineage_data.index.name = 'SampleID'
        lineage_data.to_csv(output_lineage_file, sep='\t', index=True)
        # output file three: matched mutations
        if mutation:
            mutation_data.index.name = 'SampleID'
            mutation_data['AllMutations'] = all_mutation_list
            mutation_data.to_csv(output_mutation_file, sep='\t', index=True)

        if len(no_hap_inds) != 0:
            if len(no_hap_inds) <= 5:
                self.logger.info('[Y-LineageTracker] [Warning] %s cannot classify haplogroup due to few or mismatched sites' % ', '.join(no_hap_inds))
            else:
                self.logger.info('[Y-LineageTracker] [Warning] %d samples cannot classify haplogroup due to few or mismatched sites' % len(no_hap_inds))

        self.logger.info('[Y-LineageTracker] Haplogroup classification done')

        hap_data.reset_index(level=0, inplace=True)

        return hap_data



import os
import re
import sys
import time
import logging
import argparse
import tempfile
import subprocess
import numpy as np

from ete3 import Tree


'''
Description:
This module is used for estimating divergence time of NRY haplogroups.
PAML mcmctree is wrapped in Y-LineageTracker to run time estimation analysis by Bayesian MCMC algorithm.
The tree file containing haplogroup nodes can be generated by phylo command.
'''


def time_parser():

    # function to define filter argument
    def filter_type(x):
        x = float(x)
        if x < 0 or x > 1:
            raise argparse.ArgumentTypeError('Cutoff of missing rate should be a float between 0 and 1')
        return x

    parser = argparse.ArgumentParser('time', description='(c) Y-LineageTracker: Time estimation for NRY haplogroups')
    # function used for time estimation
    parser.add_argument('time',
                        help='Estimate divergence time of NRY haplogroups.')
    # required, sequence alignment file
    parser.add_argument('--seq',
                        required=True,
                        type=str,
                        action='store',
                        help='seq: The sequence alignment file used for analysis')
    # optional, format of sequence alignment
    parser.add_argument('--seq-format',
                        required=False,
                        type=str,
                        dest='format',
                        action='store',
                        choices=['fasta', 'phylip', 'nexus', 'meg', 'vcf'],
                        help='seq-format: The format of sequence alignment file.')
    # required, tree file in newick format
    parser.add_argument('--tree',
                        required=True,
                        type=str,
                        action='store',
                        help='tree: Tree file of NRY phylogeny. \
                             The tree should be a rooted bifurcating tree without branch length. \
                             It is recommended to annotate NRY haplogroup name in tree nodes.')
    # required, calibration information of haplogroups
    parser.add_argument('--cal',
                        required=False,
                        type=str,
                        dest='calibration',
                        action='store',
                        help='cal: Haplogroup calibration information. ')
    # optional, the substitution mode
    parser.add_argument('--model',
                        required=False,
                        type=str,
                        action='store',
                        default='hky85',
                        choices=['hky85', 'jc96', 'k80', 'f81', 'f84', 'gtr'],
                        help='model: the substitution model used for time estimation')
    # optional, mutation rate of NRY
    parser.add_argument('--mut-rate',
                        required=False,
                        type=float,
                        dest='rate',
                        action='store',
                        default=7.6e-10,
                        help='mut-rate: The general mutation rate across the NRY. The default value is 7.6e-10')
    # optional, set calibration information automatically
    parser.add_argument('--auto-cal',
                        required=False,
                        dest='auto',
                        action='store_true',
                        help='auto-cal: Automatically set calibration information for the root node. \
                              In order to make this option feasible, the root node should be a haplogroup in the main trunk of the NRY tree.')
    # optional, filter sites by misisng rate value
    parser.add_argument('--filter',
                        type=filter_type,
                        required=False,
                        action='store',
                        default=0.2,
                        help='filter: Filter sites with a missing rate value greater than a value in sequence alignment file.')
    # optional, mcmc step
    parser.add_argument('--mcmc',
                        type=int,
                        required=False,
                        action='store',
                        help='mcmc: Set iteration step to run MCMC. The default is 5000.')
    # optional, the prefix of output
    parser.add_argument('-o', '--output',
                        required=False,
                        type=str,
                        action='store',
                        help='output: The prefix of output files.')

    args = parser.parse_args()

    return args


# print program information and write to log file
def set_log(log_file, args_log):

    logger = logging.getLogger()
    logger.setLevel(level=logging.INFO)

    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] - [%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    logger.addHandler(handler)
    logger.addHandler(console)

    log_info = ['[Y-LineageTracker] [Time]',
                '[Y-LineageTracker] Run Date: ' + time.asctime(time.localtime(time.time())),
                '[Y-LineageTracker] Tree File: %s' % args_log.tree,
                '[Y-LineageTracker] Sequence File: %s' % args_log.seq,
                '[Y-LineageTracker] Sequence File Format: %s' % args_log.format,
                '[Y-LineageTracker] Calibration Information: %s' % args_log.calibration,
                '[Y-LineageTracker] Mutation Rate: %s' % args_log.rate]

    if args_log.auto:
        log_info.append('[Y-LineageTracker] Perform Automatical Calibration')

    print('\n')
    for i in log_info:
        logger.info(i)


class EstimateTime(object):
    '''
    This class is used to estimate divergence time of NRY haplogroups
    PAML is wrapped in the program
    The program will generate control file of PAML mcmctree and perform analysis to estimate time
    After mcmctree is done, the program will summarize output result of mcmctree
    Three file is required for mcmctree analysis:
    1. sequence alignment file
    2. tree file in newick format
    3. calibration information file
    The tree should be a rooted bifurcating tree without branch length
    It is recommended to annotate NRY haplogroup name in tree nodes.
    '''

    def __init__(self, path):

        self.logger = logging.getLogger()
        self.path = path

    # read and clean sequence alignment file
    def _clean_alignment(self, seq_file, seq_format, filter):

        from ProcessData import ConvertData
        seq_tmp_file = tempfile.NamedTemporaryFile(mode='w')
        convert = ConvertData.ConvertData(seq_file)
        msa = convert.data_convert(seq_format, 'msa')
        seqs = [i.seq for i in msa]
        ids = [i.id for i in msa]

        if filter:
            new_seqs = []
            for i in range(len(seqs[0])):
                all_al = [s[i] for s in seqs]
                al = list(set(all_al))
                N_count = 0
                if 'N' in al:
                    N_count += all_al.count('N')
                if '-' in al:
                    N_count += all_al.count('-')
                N_rate = N_count / len(all_al)
                if N_rate < filter:
                    new_seqs.append(all_al)
            new_seqs = np.array(new_seqs).T.tolist()
        else:
            new_seqs = seqs

        max_length = max([len(i) for i in ids])
        with open(seq_tmp_file.name, 'w') as seq_tmp:
            seq_tmp.write('%d %d\n' % (len(ids), len(new_seqs[0])))
            for seq, id in zip(new_seqs, ids):
                seq_tmp.write(id+(max_length-len(id))*' '+'   '+''.join(seq)+'\n')

        return seq_tmp_file

    # read tree file
    def _read_tree_file(self, tree_file):

        tree_tmp_file = tempfile.NamedTemporaryFile(mode='w')
        tree = Tree(open(tree_file, 'r').read(), format=8)
        pruned_tree = Tree(tree.write(format=9), format=9)

        for node in pruned_tree.iter_descendants('preorder'):
            if (not node.is_leaf()) and len(node.children) == 1 and node.name == '':
                node.delete()
        taxa_num = len([leaf for leaf in pruned_tree])

        tmp_tree_info = pruned_tree.write(format=9)
        with open(tree_tmp_file.name, 'w') as tree_tmp:
            tree_tmp.write('%d 1\n\n%s' % (taxa_num, tmp_tree_info))

        return tree_tmp_file

    # read calibration information file
    def _get_calibration(self, calibration):

        if os.path.isfile(calibration):
            return open(calibration).readlines()
        else:
            return [calibration]

    # get calibration time from calibration information file
    def _haplogroup_calibration(self, calibration, tree_file):

        calibration_info = self._get_calibration(calibration)
        tree_tmp_file = tempfile.NamedTemporaryFile(mode='w')
        cal_tree = Tree(open(tree_file, 'r').read(), format=8)
        root_name = cal_tree.get_tree_root().name
        root_age = None

        used_nodes = []
        for i in calibration_info:
            used_node = i.split(':')[0]
            if re.match(r'^\w+:\d+$', i):
                cal_node, cal_time = i.split(':')
                cal_time = 'U(%s)' % cal_time
            elif re.match(r'^\w+:<\d+$', i):
                cal_node, cal_time = i.split(':<')
                cal_time = 'U(%s)' % cal_time
            elif re.match(r'^\w+:>\d+$', i):
                cal_node, cal_time = i.split(':>')
                cal_time = 'L(%s)' % cal_time
            elif re.match(r'^\w+:\d+-\d+$', i):
                cal_node, cal_time = i.split(':')
                cal_time = cal_time.split('-')[0] + ', ' + cal_time.split('-')[1]
                cal_time = 'B(%s)' % cal_time
            else:
                raise TypeError('format of calibration is incorrect')
            if used_node == root_name:
                root_age = cal_time
            else:
                try:
                    cal_tree.search_nodes(name=cal_node)[0].name = cal_time
                except:
                    raise TypeError('cannot find the haplogroup node in tree')
            used_nodes.append(used_node)

        for node in cal_tree.iter_descendants('preorder'):
            if (not node.is_leaf()):
                if len(node.children) == 1 and node.name == '':
                    node.delete()
                    continue
                if node.name not in used_nodes:
                    node.name = 'NoName'
        taxa_num = len([leaf for leaf in cal_tree])

        tmp_tree_info = cal_tree.write(format=8)
        tmp_tree_info = tmp_tree_info.replace('NoName', '')
        with open(tree_tmp_file.name, 'w') as tree_tmp:
            tree_tmp.write('%d 1\n\n%s' % (taxa_num, tmp_tree_info))

        return tree_tmp_file, root_age

    # calculate mutation rate for mcmc control file
    def _calculate_mutation_rate(self, rate):

        alpha = 1 # shape
        beta = alpha / rate
        alpha_d = 1

        return str(alpha)+' '+str(beta)+' '+str(alpha_d)

    # match model to mcmc control file
    def _get_model(self, name):

        if name == 'jc96':
            return '0'
        elif name == 'k80':
            return '1'
        elif name == 'f81':
            return '2'
        elif name == 'f84':
            return '3'
        elif name == 'hky85':
            return '4'
        elif name == 'gtr':
            return '7'

    # automatically get calibration time
    def _get_root_age(self, tree_file_name):

        from FilesIO import CommonData
        tree = Tree(open(tree_file_name, 'r').read(), format=8)
        root_name = tree.get_tree_root().name
        common_data = CommonData()
        prior_calibration = common_data.read_prior_calibration()
        if root_name in prior_calibration.index:
            lower = prior_calibration.at[root_name, 'Lower1']
            upper = prior_calibration.at[root_name, 'Upper1']
            calibration = 'B(%s, %s)' % (lower, upper)
        else:
            if root_name == '':
                self.logger.warning('[Y-LineageTracker] The root name does not exist, auto prior calibration will not be set')
            else:
                self.logger.warning('[Y-LineageTracker] Cannot find calibration time for root name %s, auto prior calibration will not be set' % root_name)
            calibration = None

        return calibration

    # get mcmc control file
    def _generate_mcmc_ctl(self, arguments, tree_file, seq_file, auto_calibration):

        if arguments.mcmc:
            if arguments.mcmc > 1000:
                mcmc = arguments.mcmc
            else:
                self.logger.warning('[Y-LineageTracker] Step for MCMC is too small, default value will be used')
                mcmc = 10000
        else:
            mcmc = 10000

        ctl_parameters={
        'seed': '-1',
        'seqfile': seq_file.name,
        'treefile': tree_file.name,
        'mcmcfile': self.path+'.mcmc.txt',
        'outfile': self.path+'.out.txt',
        'ndata': '1',
        'seqtype': '0',
        'usedata': '1',
        'clock': '1',
        'model': self._get_model(arguments.model),
        'alpha': '0',
        'ncatG': '5',
        'cleandata': '0',
        'BDparas': '1 1 0.1',
        'kappa_gamma': '6 2',
        'alpha_gamma': '1 1',
        'rgene_gamma': self._calculate_mutation_rate(arguments.rate),
        'print': '1',
        'burnin': '2000',
        'sampfreq': '5',
        'nsample': str(mcmc)}

        if auto_calibration:
            ctl_parameters['RootAge'] = auto_calibration

        max_length = max([len(i) for i in ctl_parameters.keys()])
        ctl_file = tempfile.NamedTemporaryFile(mode='w')
        with open(ctl_file.name, 'w') as ctl:
            for i in ctl_parameters.keys():
                length = len(i)
                ctl.write((max_length-length)*' '+i+' = '+ctl_parameters[i]+'\n')

        return ctl_file

    # summarize time from mcmc output
    def _summarize_result(self, orginal_tree):

        tree = Tree(orginal_tree, format=8)

        out_info = open(self.path+'.out.txt', 'r').readlines()
        for i in out_info:
            if i.startswith('Posterior mean'):
                head_num = out_info.index(i)
        time_info = out_info[head_num+2: -3]

        with open(self.path+'.haptime.txt', 'w') as time_file:
            time_file.write('NodeName\tNodeTime\tLower\tUpper\n')
            num = 0
            no_name_num = 0
            for node in tree.traverse('preorder'):
                children = node.children
                if not node.is_leaf() and len(children) > 1:
                    if node.name == '':
                        node_name = 'NoName' + str(no_name_num)
                        no_name_num += 1
                    else:
                        node_name = node.name
                    node_time_info = [i for i in time_info[num].strip().split(' ') if i != '']
                    node_time = node_time_info[1]
                    lower = node_time_info[4].strip('(').strip(',')
                    upper = node_time_info[5].strip(')')
                    num += 1
                    time_file.write(node_name+'\t'+node_time+'\t'+lower+'\t'+upper+'\n')

    # main function to run time estimation
    def run_mcmc(self, arguments):

        if not (arguments.auto or arguments.calibration):
            self.logger.error('[Y-LineageTracker] Stopped since no calibration information')
            sys.exit()

        # get calibration time
        self.logger.info('[Y-LineageTracker] Read tree file and calibration')
        if arguments.auto:
            auto_calibration = self._get_root_age(arguments.tree)
        else:
            auto_calibration = None
        if arguments.calibration:
            tree_file, auto_calibration = self._haplogroup_calibration(arguments.calibration, arguments.tree)
        else:
            tree_file = self._read_tree_file(arguments.tree)

        # get sequence alignment file
        self.logger.info('[Y-LineageTracker] Read seq file')
        seq_file = self._clean_alignment(arguments.seq, arguments.format, arguments.filter)

        # get control file
        ctl_file = self._generate_mcmc_ctl(arguments, tree_file, seq_file, auto_calibration)

        # get control file and run mcmctree
        self.logger.info('[Y-LineageTracker] Estimating divergence time, this may take some time')
        mcmc = os.path.split(os.path.realpath(__file__))[0] + '/src/mcmctree'
        cmd = mcmc + ' ' + ctl_file.name
        FNULL = open(os.devnull, 'w')
        process = subprocess.Popen(cmd, shell=True, stdout=FNULL, stderr=subprocess.STDOUT)
        process.wait()

        time.sleep(0.1)

        # summarize output result
        self.logger.info('[Y-LineageTracker] Summarizing output result')
        self._summarize_result(arguments.tree)


def main():

    start = time.perf_counter()
    arguments = time_parser()

    from FilesIO import time_count, get_out_path
    # get path of output files
    path = get_out_path(arguments.output, 'TimeEstimation')
    log_file = path + '.TimeLog.log'

    # set log file
    set_log(log_file, arguments)

    # estimate divergence time
    estimate = EstimateTime(path)
    estimate.run_mcmc(arguments)

    time_count(start)


if __name__ == '__main__':
    main()
